"""

Compare parameters
==================

Function to compare estimated parameters with true parameters.

:author: Michel Bierlaire
:date: Wed Nov  1 18:07:08 2023
"""
import pandas as pd
from biogeme.results import bioResults
from true_parameters import true_parameters


# %%
def compare(estimated_parameters: bioResults) -> tuple[pd.DataFrame, str]:
    """Compare estimated and true parameters

    :param estimated_parameters: estimation results
    """
    non_estimated = []
    data = []
    for name, value in true_parameters.items():
        try:
            est_value = estimated_parameters.at[name, 'Value']
            std_err = estimated_parameters.at[name, 'Rob. Std err']
            t_test = (value - est_value) / std_err
            # Append the data to the list instead of printing
            data.append(
                {
                    'Name': name,
                    'True Value': value,
                    'Estimated Value': est_value,
                    'T-Test': t_test,
                }
            )
        except KeyError:
            non_estimated.append(name)

    # Convert the list of dictionaries to a DataFrame

    msg = f'Parameters not estimated: {non_estimated}' if non_estimated else ''
    df = pd.DataFrame(data)
    return df, msg



"""Function renaming variables inside an expression

Michel Bierlaire
Fri May 02 2025, 13:51:23
"""

from .base_expressions import Expression
from .variable import Variable
from .visitor import ExpressionVisitor

_rename_visitor = ExpressionVisitor()
register_rename = _rename_visitor.register


@register_rename(Variable)
def rename_variable_handler(expr, context):
    if expr.name == context['old_name']:
        expr.name = context['new_name']
        context['count'] += 1


def rename_all_variables(expr: Expression, old_name: str, new_name: str) -> int:
    context = {'old_name': old_name, 'new_name': new_name, 'count': 0}
    _rename_visitor.visit(expr, context)
    return context['count']



""".. _plot_b21process_pareto:

Re-estimate the Pareto optimal models
=====================================

The assisted specification algorithm generates a file containg the
pareto optimal specification. This script is designed to re-estimate
the Pareto optimal models. The catalog of specifications is defined in
:ref:`plot_b21multiple_models_spec` .

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 17:46:14 2023

"""
import biogeme.biogeme_logging as blog

try:
    import matplotlib.pyplot as plt

    can_plot = True
except ModuleNotFoundError:
    can_plot = False
from biogeme_optimization.exceptions import OptimizationError
from biogeme.assisted import ParetoPostProcessing
from biogeme.results import compileEstimationResults
from plot_b21multiple_models_spec import the_biogeme

PARETO_FILE_NAME = 'saved_results/b21multiple_models.pareto'

logger = blog.get_screen_logger(blog.INFO)
logger.info('Example b21process_pareto.py')

CSV_FILE = 'b21process_pareto.csv'
SEP_CSV = ','

# %%
# The constructor of the Pareto post processing object takes two arguments:
#
#    - the biogeme object,
#    - the name of the file where the algorithm has stored the
#      estimated models.
the_pareto_post = ParetoPostProcessing(
    biogeme_object=the_biogeme,
    pareto_file_name=PARETO_FILE_NAME,
)

# %%
the_pareto_post.log_statistics()

# %%
# Complete re-estimation of the best models, including the calculation
# of the statistics.
all_results = the_pareto_post.reestimate(recycle=False)

# %%
summary, description = compileEstimationResults(all_results, use_short_names=True)
print(summary)

# %%
print(f'Summary table available in {CSV_FILE}')
summary.to_csv(CSV_FILE, sep=SEP_CSV)

# %%
# Explanation of the short names of the models.
with open(CSV_FILE, 'a', encoding='utf-8') as f:
    print('\n\n', file=f)
    for k, v in description.items():
        if k != v:
            print(f'{k}: {v}')
            print(f'{k}{SEP_CSV}{v}', file=f)

# %%
# The following plot illustrates all models that have been estimated.
# Each dot corresponds to a model. The x-coordinate corresponds to the
# negative log-likelihood. The y-coordinate corresponds to the number
# of parameters. If the shape of the dot is a circle, it means that it
# corresponds to a Pareto optimal model. If the shape is a cross, it
# means that the model has been Pareto optimal at some point during
# the algorithm and later removed as a new model dominating it has
# been found.
if can_plot:
    try:
        _ = the_pareto_post.plot(
            label_x='Negative loglikelihood', label_y='Number of parameters'
        )
        plt.show()
    except OptimizationError as e:
        print(f'No plot available: {e}')



"""
Test the function creating LaTeX code

Michel Bierlaire
Tue Oct 1 19:18:45 2024
"""

import unittest

from biogeme.results_processing import latex_output
from biogeme.results_processing.latex_output import get_sign_for_p_value


class TestAddTrailingZero(unittest.TestCase):

    def test_string_without_period(self):
        """Test case for numbers without a period (should append .0)."""
        self.assertEqual(latex_output.add_trailing_zero("123"), "123.0")
        self.assertEqual(latex_output.add_trailing_zero("456"), "456.0")
        self.assertEqual(latex_output.add_trailing_zero("0"), "0.0")

    def test_string_with_period(self):
        """Test case for numbers already containing a period (should remain unchanged)."""
        self.assertEqual(latex_output.add_trailing_zero("123.45"), "123.45")
        self.assertEqual(latex_output.add_trailing_zero("0.123"), "0.123")
        self.assertEqual(latex_output.add_trailing_zero("100."), "100.0")

    def test_empty_string(self):
        """Test case for empty string input (should append .0)."""
        self.assertEqual(latex_output.add_trailing_zero(""), "0.0")


class TestFormatRealNumber(unittest.TestCase):

    def test_non_scientific_format(self):
        # Test values that don't need scientific notation
        self.assertEqual(latex_output.format_real_number(123.456), '123.0')
        self.assertEqual(latex_output.format_real_number(0.0123456), '0.0123')
        self.assertEqual(latex_output.format_real_number(999.999), '1.0e+03')

    def test_scientific_lowercase_e(self):
        # Test values that result in scientific notation with 'e'
        self.assertEqual(latex_output.format_real_number(1.2345e-5), '1.23e-05')
        self.assertEqual(latex_output.format_real_number(1e-10), '1.0e-10')
        self.assertEqual(latex_output.format_real_number(3.1415e20), '3.14e+20')

    def test_scientific_uppercase_E(self):
        # Test values that result in scientific notation with 'E'
        # If necessary, you can modify format_real_number to support 'E' if required
        self.assertEqual(latex_output.format_real_number(2.5e-4), '0.00025')
        self.assertEqual(latex_output.format_real_number(9.8765e10), '9.88e+10')

    def test_edge_cases(self):
        # Test very small and very large numbers
        self.assertEqual(latex_output.format_real_number(1e-100), '1.0e-100')
        self.assertEqual(latex_output.format_real_number(1e100), '1.0e+100')

    def test_zero_and_near_zero(self):
        # Test zero and very small values close to zero
        self.assertEqual(latex_output.format_real_number(0), '0.0')
        self.assertEqual(latex_output.format_real_number(0.00000001), '1.0e-08')

    def test_negative_numbers(self):
        # Test negative values
        self.assertEqual(latex_output.format_real_number(-123.456), '-123.0')
        self.assertEqual(latex_output.format_real_number(-0.000123), '-0.000123')
        self.assertEqual(latex_output.format_real_number(-1.2345e-5), '-1.23e-05')

    def test_trailing_zeros(self):
        # Test proper formatting of numbers that require trailing zeros
        self.assertEqual(latex_output.format_real_number(1), '1.0')
        self.assertEqual(latex_output.format_real_number(10), '10.0')
        self.assertEqual(latex_output.format_real_number(1.234), '1.23')


class TestGetSignForPValue(unittest.TestCase):
    def test_valid_thresholds(self):
        p_thresholds = [(0.01, '***'), (0.05, '**'), (0.1, '*'), (0.2, '+')]
        self.assertEqual(
            get_sign_for_p_value(0.005, p_thresholds), "***"
        )  # Smallest p-value
        self.assertEqual(
            get_sign_for_p_value(0.03, p_thresholds), "**"
        )  # Between 0.01 and 0.05
        self.assertEqual(
            get_sign_for_p_value(0.07, p_thresholds), "*"
        )  # Between 0.05 and 0.1
        self.assertEqual(
            get_sign_for_p_value(0.15, p_thresholds), "+"
        )  # Between 0.1 and 0.2
        self.assertEqual(get_sign_for_p_value(0.25, p_thresholds), "")  # No match

    def test_string(self):
        p_thresholds = [(0.01, '***'), (0.05, '**'), (0.1, '*'), (0.2, '+')]
        self.assertEqual(
            get_sign_for_p_value('0.005', p_thresholds), "***"
        )  # Smallest p-value
        p_thresholds = [(0.01, 12), (0.05, '**'), (0.1, '*'), (0.2, '+')]
        self.assertEqual(
            get_sign_for_p_value(0.005, p_thresholds), '12'
        )  # Smallest p-value

    def test_exact_threshold_matches(self):
        p_thresholds = [(0.01, "***"), (0.05, "**"), (0.1, "*"), (0.2, "+")]
        self.assertEqual(get_sign_for_p_value(0.01, p_thresholds), "***")
        self.assertEqual(get_sign_for_p_value(0.05, p_thresholds), "**")
        self.assertEqual(get_sign_for_p_value(0.1, p_thresholds), "*")
        self.assertEqual(get_sign_for_p_value(0.2, p_thresholds), "+")

    def test_empty_threshold_list(self):
        self.assertEqual(get_sign_for_p_value(0.05, []), "")  # No thresholds given

    def test_invalid_p_value(self):
        p_thresholds = [(0.01, "***"), (0.05, "**"), (0.1, "*")]

        with self.assertRaises(TypeError):
            get_sign_for_p_value(None, p_thresholds)  # None instead of float

        with self.assertRaises(TypeError):
            get_sign_for_p_value('a_string', p_thresholds)  # None instead of float

    def test_invalid_threshold_structure(self):
        with self.assertRaises(TypeError):
            get_sign_for_p_value(0.05, [0.01, 0.05, 0.1])  # Not a list of tuples
        with self.assertRaises(TypeError):
            get_sign_for_p_value(
                0.05, [(0.01, "***"), ("invalid", "**")]
            )  # Non-float threshold

    def test_unsorted_thresholds(self):
        p_thresholds = [(0.1, "*"), (0.01, "***"), (0.05, "**")]  # Unsorted list
        self.assertEqual(
            get_sign_for_p_value(0.03, p_thresholds), "**"
        )  # Finds correct minimum threshold


if __name__ == "__main__":
    unittest.main()



from .sampling_context import (
    StratumTuple,
    CrossVariableTuple,
    SamplingContext,
    MEV_PREFIX,
    MEV_WEIGHT,
    LOG_PROBA_COL,
)
from .choice_set_generation import ChoiceSetsGeneration
from .sampling_of_alternatives import generate_segment_size, SamplingOfAlternatives
from .generate_model import GenerateModel



"""Iterator on arithmetic expressions in a catalog

Michel Bierlaire
Thu Apr 17 2025, 08:30:58
"""

from __future__ import annotations

from biogeme.exceptions import BiogemeError

from .central_controller import CentralController
from .configuration import Configuration


class SelectedConfigurationsIterator:
    """A multiple expression is an expression that contains
    Catalog. This iterator loops on pre-specified configurations
    """

    def __init__(
        self,
        the_central_controller: CentralController,
        selected_configurations: set[Configuration] | None = None,
    ):
        """Ctor.

        :param the_central_controller: expression containing Catalogs
        :param selected_configurations: selected configurations to iterate on. If None, all configurations are considered.
        """
        self.the_central_controller = the_central_controller
        self.configurations = (
            selected_configurations or the_central_controller.all_configurations
        )
        if not isinstance(self.configurations, set):
            error_msg = f'The selected configurations must be a set, and not an object of type {type(self.configurations)}'
            raise BiogemeError(error_msg)
        self.set_iterator = iter(self.configurations)
        self.current_configuration = next(self.set_iterator)
        self.the_central_controller.set_configuration(self.current_configuration)
        self.first = True
        self.number = 0

    def __iter__(self) -> SelectedConfigurationsIterator:
        return self

    def __next__(self) -> Configuration:
        self.number += 1
        if self.first:
            self.first = False
            return self.current_configuration

        self.current_configuration = next(self.set_iterator)
        self.the_central_controller.set_configuration(self.current_configuration)
        return self.current_configuration



"""

Simple integral
===============

Calculation of a simple integral using Monte-Carlo integration.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 20:42:24 2023

"""

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme.expressions import exp, bioDraws, MonteCarlo
from biogeme.tools import TemporaryFile

# %%
# We create a fake database with one entry, as it is required
# to store the draws
df = pd.DataFrame()
df['FakeColumn'] = [1.0]
database = db.Database('fakeDatabase', df)

# %%
integrand = exp(bioDraws('U', 'UNIFORM'))
simulatedI = MonteCarlo(integrand)

# %%
trueI = exp(1.0) - 1.0

# %%
R = 200
MULTIPLIER = 100000

# %%
# Create a parameter file to set the number of draws.
with TemporaryFile() as filename:
    with open(filename, 'w', encoding='utf-8') as f:
        print('[MonteCarlo]', file=f)
        print(f'number_of_draws = {R}', file=f)

    sampleVariance = MonteCarlo(integrand * integrand) - simulatedI * simulatedI
    stderr = (sampleVariance / R) ** 0.5
    error = simulatedI - trueI

    simulate = {
        'Analytical Integral': trueI,
        'Simulated Integral': simulatedI,
        'Sample variance   ': sampleVariance,
        'Std Error         ': stderr,
        'Error             ': error,
    }

    biosim = bio.BIOGEME(database, simulate, parameter_file=filename)
    R = biosim.number_of_draws
    biosim.modelName = f'01simpleIntegral_{R}'
    results = biosim.simulate(theBetaValues={})
    print(f'Number of draws: {R}')
    for c in results.columns:
        print(f'{c}: {results.loc[0,c]}')

# %%
# Create a parameter file to set the number of draws.
with TemporaryFile() as filename:
    with open(filename, 'w', encoding='utf-8') as f:
        print('[MonteCarlo]', file=f)
        print(f'number_of_draws = {MULTIPLIER * R}', file=f)

    biogeme2 = bio.BIOGEME(database, simulate, parameter_file=filename)
    biogeme2.modelName = '01simpleIntegral_{multiplier*R}'
    results2 = biogeme2.simulate(theBetaValues={})
    print(f'Number of draws: {MULTIPLIER * R}')
    for c in results2.columns:
        print(f'{c}: {results2.loc[0, c]}')



"""

biogeme.version
===============

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Thu Dec  7 15:22:06 2023
"""

import biogeme.version as ver

# %%
# Obtain the version number
ver.getVersion()

# %%
# Package information in text format.
print(ver.getText())

# %%
# Package information in HTML format
print(ver.getHtml())

# %%
# Package information in LaTeX format
print(ver.getLaTeX())

# %%
# Defines the `__version__` variable.
print(ver.__version__)



import unittest
from datetime import timedelta

from biogeme.results_processing import RawEstimationResults
from biogeme.results_processing.compilation import compile_estimation_results
from biogeme.results_processing.estimation_results import EstimationResults


class TestCompileEstimationResults(unittest.TestCase):

    def setUp(self):
        raw_results = RawEstimationResults(
            model_name='test',
            user_notes='',
            beta_names=['beta1', 'beta2'],
            beta_values=[1.0, 2.0],
            lower_bounds=[-100, -100],
            upper_bounds=[100, 100],
            gradient=[0, 0],
            hessian=[[1, 1], [1, 1]],
            bhhh=[[1, 1], [1, 1]],
            null_log_likelihood=-1000,
            initial_log_likelihood=-1000,
            final_log_likelihood=-100,
            data_name='data',
            sample_size=500,
            number_of_observations=500,
            monte_carlo=False,
            number_of_draws=0,
            types_of_draws={},
            number_of_excluded_data=0,
            draws_processing_time=timedelta(1),
            optimization_messages={},
            convergence=True,
            bootstrap=[],
            bootstrap_time=None,
        )
        # Create a real instance of EstimationResults
        self.mock_results = EstimationResults(raw_results)

    def test_compile_estimation_results_basic(self):
        """Test basic functionality of compile_estimation_results."""
        dict_of_results = {
            'Model_1': self.mock_results,
            'Model_2': self.mock_results,
        }

        df, config = compile_estimation_results(
            dict_of_results=dict_of_results,
            statistics=('Number of estimated parameters', 'Sample size'),
            include_parameter_estimates=False,
            include_stderr=False,
            include_t_test=False,
            formatted=False,
            use_short_names=True,
        )

        # Verify the dataframe structure
        self.assertIn('Model_000000', df.columns)
        self.assertIn('Model_000001', df.columns)
        self.assertIn('Number of estimated parameters', df.index)
        self.assertIn('Sample size', df.index)

        # Verify the values
        self.assertEqual(df.loc['Number of estimated parameters', 'Model_000000'], '2')
        self.assertEqual(df.loc['Sample size', 'Model_000001'], '500')

    def test_compile_estimation_results_with_parameters(self):
        """Test compile_estimation_results with parameter estimates, stderr, and t-test."""
        dict_of_results = {
            'Model_1': self.mock_results,
        }

        df, config = compile_estimation_results(
            dict_of_results=dict_of_results,
            statistics=('Final log likelihood',),
            include_parameter_estimates=True,
            include_stderr=True,
            include_t_test=True,
            formatted=True,
            use_short_names=False,
        )

        # Check the inclusion of parameters, standard errors, and t-tests in formatted results
        self.assertIn('beta1 (std) (t-test)', df.index)
        self.assertIn('beta2 (std) (t-test)', df.index)
        self.assertEqual(df.loc['beta1 (std) (t-test)', 'Model_1'], '1 (0.5) (2)')
        self.assertEqual(df.loc['beta2 (std) (t-test)', 'Model_1'], '2 (0.5) (4)')


if __name__ == '__main__':
    unittest.main()



"""Unit tests for negative loglikelihood function

Michel Bierlaire
Sat Mar 29 16:31:28 2025
"""

import unittest

import numpy as np
from biogeme_optimization.function import FunctionData

from biogeme.exceptions import BiogemeError
from biogeme.function_output import FunctionOutput
from biogeme.likelihood.negative_likelihood import NegativeLikelihood


class TestNegativeLikelihood(unittest.TestCase):
    def setUp(self):
        self.dimension = 2
        self.mock_output = FunctionOutput(
            function=10.0,
            gradient=np.array([1.0, 2.0]),
            hessian=np.array([[1.0, 0.0], [0.0, 3.0]]),
            bhhh=None,
        )

        def mock_loglikelihood(x, gradient, hessian, bhhh):
            return self.mock_output

        self.neg_likelihood = NegativeLikelihood(self.dimension, mock_loglikelihood)
        self.neg_likelihood.x = np.array([1.0, 2.0])

    def test_dimension(self):
        self.assertEqual(self.neg_likelihood.dimension(), self.dimension)

    def test_f_returns_negative_loglikelihood(self):
        result = self.neg_likelihood._f()
        self.assertAlmostEqual(result, -10.0)

    def test_f_g_returns_negative_function_and_gradient(self):
        result: FunctionData = self.neg_likelihood._f_g()
        self.assertAlmostEqual(result.function, -10.0)
        np.testing.assert_array_almost_equal(
            result.gradient, -self.mock_output.gradient
        )
        self.assertIsNone(result.hessian)

    def test_f_g_h_returns_negative_function_gradient_and_hessian(self):
        result: FunctionData = self.neg_likelihood._f_g_h()
        self.assertAlmostEqual(result.function, -10.0)
        np.testing.assert_array_almost_equal(
            result.gradient, -self.mock_output.gradient
        )
        np.testing.assert_array_almost_equal(result.hessian, -self.mock_output.hessian)

    def test_f_raises_error_if_x_not_set(self):
        self.neg_likelihood.x = None
        with self.assertRaises(BiogemeError):
            self.neg_likelihood._f()

    def test_f_g_raises_error_if_x_not_set(self):
        self.neg_likelihood.x = None
        with self.assertRaises(BiogemeError):
            self.neg_likelihood._f_g()

    def test_f_g_h_raises_error_if_x_not_set(self):
        self.neg_likelihood.x = None
        with self.assertRaises(BiogemeError):
            self.neg_likelihood._f_g_h()


if __name__ == '__main__':
    unittest.main()



""" Implements the cross-nested logit model.

:author: Michel Bierlaire
:date: Wed Oct 25 11:08:59 2023
"""

import logging

from biogeme.deprecated import deprecated
from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Expression,
    exp,
    MultipleSum,
    logzero,
    log,
    ExpressionOrNumeric,
)
from biogeme.models import logmev
from biogeme.nests import NestsForCrossNestedLogit, OldNestsForCrossNestedLogit

logger = logging.getLogger(__name__)


def cnl(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForCrossNestedLogit | OldNestsForCrossNestedLogit,
    choice: ExpressionOrNumeric,
) -> Expression:
    """Implements the cross-nested logit model as a MEV model.

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: object describing the nesting structure

    :param choice: id of the alternative for which the probability must be
              calculated.

    :return: choice probability for the cross-nested logit model.
    """
    return exp(logcnl(util, availability, nests, choice))


@deprecated(cnl)
def cnl_avail(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric],
    nests: NestsForCrossNestedLogit | OldNestsForCrossNestedLogit,
    choice: ExpressionOrNumeric,
) -> Expression:
    """Same as cnl. Maintained for backward compatibility"""
    pass


@deprecated(cnl)
def logcnl_avail(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForCrossNestedLogit | OldNestsForCrossNestedLogit,
    choice: ExpressionOrNumeric,
) -> Expression:
    """Same as logcnl. Maintained for backward compatibility

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: object describing the nesting structure

    :param choice: id of the alternative for which the probability must be
              calculated.

    :return: log of choice probability for the cross-nested logit model.
    """
    pass


def get_mev_for_cross_nested(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForCrossNestedLogit | OldNestsForCrossNestedLogit,
) -> dict[int, Expression]:
    """Implements the MEV generating function for the cross nested logit
    model as a MEV model.

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
        alternative, indexed
        by numerical ids. Must be consistent with util, or
        None. In this case, all alternatives are supposed to be
        always available.

    :param nests: object describing the nesting structure

    :return: log of the choice probability for the cross-nested logit model.
    """
    if not isinstance(nests, NestsForCrossNestedLogit):
        logger.warning(
            'It is recommended to define the nests of the cross-nested logit model using '
            'the objects OneNestForNestedLogit and NestsForCrossNestedLogit defined '
            'in biogeme.nests.'
        )
        nests = NestsForCrossNestedLogit(choice_set=list(util), tuple_of_nests=nests)

    ok, message = nests.check_validity()
    if not ok:
        raise BiogemeError(message)

    gi_terms: dict[int, Expression] = {}
    if nests.alone is None:
        log_gi = {}
        for i in util:
            gi_terms[i] = []
    else:
        log_gi = {i: 0 for i in nests.alone}
        for i in set(util).difference(set(nests.alone)):
            gi_terms[i] = []
    for m in nests:
        if availability is None:
            biosum = MultipleSum(
                [
                    a**m.nest_param * exp(m.nest_param * (util[i]))
                    for i, a in m.dict_of_alpha.items()
                ]
            )
        else:
            biosum = MultipleSum(
                [
                    availability[i] * a**m.nest_param * exp(m.nest_param * (util[i]))
                    for i, a in m.dict_of_alpha.items()
                ]
            )
        for i, a in m.dict_of_alpha.items():
            gi_terms[i] += [
                a**m.nest_param
                * exp((m.nest_param - 1) * (util[i]))
                * biosum ** ((1.0 - m.nest_param) / m.nest_param)
            ]
    for k, G in gi_terms.items():
        log_gi[k] = logzero(MultipleSum(G))
    return log_gi


@deprecated(new_func=get_mev_for_cross_nested)
def getMevForCrossNested(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForCrossNestedLogit | OldNestsForCrossNestedLogit,
) -> Expression:
    pass


def logcnl(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForCrossNestedLogit | OldNestsForCrossNestedLogit,
    choice: ExpressionOrNumeric,
) -> Expression:
    """Implements the log of the cross-nested logit model as a MEV model.

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: object describing the nesting structure
    :param choice: id of the alternative for which the probability must be
              calculated.

    :return: log of the choice probability for the cross-nested logit model.

    :raise BiogemeError: if the definition of the nests is invalid.
    """

    log_gi = get_mev_for_cross_nested(util, availability, nests)
    log_p = logmev(util, log_gi, availability, choice)
    return log_p


def cnlmu(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForCrossNestedLogit,
    choice: ExpressionOrNumeric,
    mu: ExpressionOrNumeric,
) -> Expression:
    """Implements the cross-nested logit model as a MEV model with
    the homogeneity parameters is explicitly involved

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: object describing the nesting structure

    :param choice: id of the alternative for which the probability must be
              calculated.

    :param mu: Homogeneity parameter :math:`\\mu`.

    :return: choice probability for the cross-nested logit model.
    """
    return exp(logcnlmu(util, availability, nests, choice, mu))


def get_mev_for_cross_nested_mu(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForCrossNestedLogit | OldNestsForCrossNestedLogit,
    mu: ExpressionOrNumeric,
) -> dict[int, Expression]:
    """Implements the MEV generating function for the cross-nested logit
    model as a MEV model with the homogeneity parameters is explicitly
    involved.

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: object describing the nesting structure

    :param mu: Homogeneity parameter :math:`\\mu`.

    :return: log of the choice probability for the cross-nested logit model.

    """

    if not isinstance(nests, NestsForCrossNestedLogit):
        logger.warning(
            'It is recommended to define the nests of the cross-nested logit model using '
            'the objects OneNestForNestedLogit and NestsForCrossNestedLogit defined '
            'in biogeme.nests.'
        )
        nests = NestsForCrossNestedLogit(choice_set=list(util), tuple_of_nests=nests)

    ok, message = nests.check_validity()
    if not ok:
        raise BiogemeError(message)

    gi_terms = {}
    if nests.alone is None:
        log_gi = {}
        for i in util:
            gi_terms[i] = []
    else:
        log_gi = {i: log(mu) + (mu - 1) * util[i] for i in nests.alone}
        for i in set(util).difference(set(nests.alone)):
            gi_terms[i] = []
    for m in nests:
        if availability is None:
            biosum = MultipleSum(
                [
                    a ** (m.nest_param / mu) * exp(m.nest_param * (util[i]))
                    for i, a in m.dict_of_alpha.items()
                ]
            )
        else:
            biosum = MultipleSum(
                [
                    availability[i]
                    * a ** (m.nest_param / mu)
                    * exp(m.nest_param * (util[i]))
                    for i, a in m.dict_of_alpha.items()
                ]
            )
        for i, a in m.dict_of_alpha.items():
            gi_terms[i] += [
                a ** (m.nest_param / mu)
                * exp((m.nest_param - 1) * (util[i]))
                * biosum ** ((mu / m.nest_param) - 1.0)
            ]
    for k, G in gi_terms.items():
        log_gi[k] = log(mu * MultipleSum(G))
    return log_gi


@deprecated(get_mev_for_cross_nested_mu)
def getMevForCrossNestedMu(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForCrossNestedLogit | OldNestsForCrossNestedLogit,
    mu: ExpressionOrNumeric,
) -> dict[int, Expression]:
    pass


def logcnlmu(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForCrossNestedLogit,
    choice: ExpressionOrNumeric,
    mu: ExpressionOrNumeric,
) -> Expression:
    """Implements the log of the cross-nested logit model as a MEV model
    with the homogeneity parameters is explicitly involved.


    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: object describing the nesting structure

    :param choice: id of the alternative for which the probability must be
              calculated.

    :param mu: Homogeneity parameter :math:`\\mu`.

    :return: log of the choice probability for the cross-nested logit model.

    :raise BiogemeError: if the definition of the nests is invalid.

    """
    log_gi = get_mev_for_cross_nested_mu(util, availability, nests, mu)
    log_p = logmev(util, log_gi, availability, choice)
    return log_p



"""

Cross-nested logit
==================

Estimation of a cross-nested logit model using sampling of alternatives.

:author: Michel Bierlaire
:date: Wed Nov  1 18:00:33 2023
"""
import pandas as pd
from biogeme.sampling_of_alternatives import (
    SamplingContext,
    ChoiceSetsGeneration,
    GenerateModel,
    generate_segment_size,
)
from biogeme.expressions import Beta
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme.nests import OneNestForCrossNestedLogit, NestsForCrossNestedLogit
from specification import V, combined_variables
from compare import compare
from alternatives import (
    alternatives,
    ID_COLUMN,
    partitions,
    all_alternatives,
    asian_and_downtown,
    only_downtown,
    only_asian,
)

# %%
logger = blog.get_screen_logger(level=blog.INFO)

# %%
PARTITION = 'downtown'
MEV_PARTITION = 'uniform_asian_or_downtown'
SAMPLE_SIZE = 10  # out of 100 alternatives
SAMPLE_SIZE_MEV = 63  # out of 63 alternatives
CHOICE_COLUMN = 'cnl_3'
MODEL_NAME = f'cnl_{SAMPLE_SIZE}_{SAMPLE_SIZE_MEV}'
FILE_NAME = f'{MODEL_NAME}.dat'

# %%
the_partition = partitions.get(PARTITION)
if the_partition is None:
    raise ValueError(f'Unknown partition: {PARTITION}')

# %%
segment_sizes = list(
    generate_segment_size(SAMPLE_SIZE, the_partition.number_of_segments())
)

# %%
# We use all alternatives in the nest.
mev_partition = partitions.get(MEV_PARTITION)
if mev_partition is None:
    raise ValueError(f'Unknown partition: {MEV_PARTITION}')
mev_segment_sizes = [
    SAMPLE_SIZE_MEV,
]

# %%
# Nests

# %%
# Downtown
mu_downtown = Beta('mu_downtown', 1, 1, None, 0)
downtown_alpha_dict = {i: 0.5 for i in asian_and_downtown} | {
    i: 1 for i in only_downtown
}
downtown_nest = OneNestForCrossNestedLogit(
    nest_param=mu_downtown, dict_of_alpha=downtown_alpha_dict, name='downtown'
)

# %%
# Asian
mu_asian = Beta('mu_asian', 1, 1, None, 0)
asian_alpha_dict = {i: 0.5 for i in asian_and_downtown} | {i: 1.0 for i in only_asian}
asian_nest = OneNestForCrossNestedLogit(
    nest_param=mu_asian, dict_of_alpha=asian_alpha_dict, name='asian'
)

cnl_nests = NestsForCrossNestedLogit(
    choice_set=all_alternatives,
    tuple_of_nests=(downtown_nest, asian_nest),
)

# %%
observations = pd.read_csv('obs_choice.dat')

# %%
context = SamplingContext(
    the_partition=the_partition,
    sample_sizes=segment_sizes,
    individuals=observations,
    choice_column=CHOICE_COLUMN,
    alternatives=alternatives,
    id_column=ID_COLUMN,
    biogeme_file_name=FILE_NAME,
    utility_function=V,
    combined_variables=combined_variables,
    mev_partition=mev_partition,
    mev_sample_sizes=mev_segment_sizes,
    cnl_nests=cnl_nests,
)

# %%
logger.info(context.reporting())

# %%
the_data_generation = ChoiceSetsGeneration(context=context)
the_model_generation = GenerateModel(context=context)

# %%
biogeme_database = the_data_generation.sample_and_merge(recycle=False)

# %%
logprob = the_model_generation.get_cross_nested_logit()

# %%
the_biogeme = bio.BIOGEME(biogeme_database, logprob)
the_biogeme.modelName = MODEL_NAME

# %%
# Calculate the null log likelihood for reporting.
the_biogeme.calculateNullLoglikelihood({i: 1 for i in range(context.total_sample_size)})

# %%
# Estimate the parameters.
results = the_biogeme.estimate(recycle=False)

# %%
print(results.short_summary())

# %%
estimated_parameters = results.getEstimatedParameters()
estimated_parameters

# %%
df, msg = compare(estimated_parameters)

# %%
print(df)

# %%
print(msg)



"""
Compilation of estimation results

Michel Bierlaire
Thu Oct 3 18:54:13 2024
"""

import glob
import logging
import os

import pandas as pd

from biogeme.results_processing.estimation_results import (
    EstimationResults,
    EstimateVarianceCovariance,
)
from biogeme.tools import ModelNames

logger = logging.getLogger(__name__)


def compile_estimation_results(
    dict_of_results: dict[str, EstimationResults | str],
    variance_covariance_type: EstimateVarianceCovariance = EstimateVarianceCovariance.ROBUST,
    statistics: tuple[str, ...] = (
        'Number of estimated parameters',
        'Sample size',
        'Final log likelihood',
        'Akaike Information Criterion',
        'Bayesian Information Criterion',
    ),
    include_parameter_estimates: bool = True,
    include_stderr: bool = False,
    include_t_test: bool = True,
    formatted: bool = True,
    use_short_names: bool = False,
) -> tuple[pd.DataFrame, dict[str, str]]:
    """Compile estimation results into a common table

    :param dict_of_results: dict of results, containing
        for each model the name, the ID and the results, or the name
        of the pickle file containing them.
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param statistics: list of statistics to include in the summary
        table
    :param include_parameter_estimates: if True, the parameter
        estimates are included.
    :param include_stderr: if True, the robust standard errors
         of the parameters are included.
    :param include_t_test: if True, the t-test
         of the parameters are included.
    :param formatted: if True, a formatted string in included in the
         table results. If False, the numerical values are stored. Use
         "True" if you need to print the results. Use "False" if you
         need to use them for further calculation.
    :param use_short_names: if True, short names, such as Model_1,
        Model_2, are used to identify the model. It is nicer on for the
        reporting.
    :return: pandas dataframe with the requested results, and a dictionary reporting the
        specification of each model

    """
    model_names = ModelNames()

    def the_name(col: str) -> str:
        """Replace the name of a model by a shorter version for reporting

        :param col: name of the column, that is, name of the model.
        :return: name to be used in the reporting.
        """
        if use_short_names:
            return model_names(col)
        return col

    columns = [the_name(k) for k in dict_of_results.keys()]
    df = pd.DataFrame(columns=columns)

    configurations = {the_name(col): col for col in dict_of_results.keys()}

    for model, estimation_results in dict_of_results.items():
        if use_short_names:
            col = model_names(model)
        else:
            col = model
        if not isinstance(estimation_results, EstimationResults):
            try:
                estimation_results = EstimationResults.from_yaml_file(
                    filename=estimation_results
                )
            except FileNotFoundError:
                warning = f'Impossible to access result file {estimation_results}'
                logger.warning(warning)
                estimation_results = None

        if estimation_results is not None:
            stats_results = estimation_results.get_general_statistics()
            for s in statistics:
                df.loc[s, col] = stats_results[s]
            if include_parameter_estimates:
                for (
                    parameter_index,
                    parameter_name,
                ) in enumerate(estimation_results.beta_names):
                    parameter_value = estimation_results.get_parameter_value_from_index(
                        parameter_index=parameter_index
                    )
                    std_err_value = estimation_results.get_parameter_std_err_from_index(
                        parameter_index=parameter_index,
                        estimate_var_covar=variance_covariance_type,
                    )
                    t_test_value = estimation_results.get_parameter_t_test_from_index(
                        parameter_index=parameter_index,
                        estimate_var_covar=variance_covariance_type,
                    )
                    if formatted:

                        std_err_report = (
                            f'({std_err_value:.3g})' if include_stderr else ''
                        )
                        t_test_report = (
                            f'({t_test_value:.3g})' if include_t_test else ''
                        )
                        the_value = (
                            f'{parameter_value:.3g} {std_err_report} {t_test_report}'
                        )
                        row_std = ' (std)' if include_stderr else ''
                        row_t_test = ' (t-test)' if include_t_test else ''
                        row_title = f'{parameter_name}{row_std}{row_t_test}'
                        df.loc[row_title, col] = the_value
                    else:
                        df.loc[parameter_name, col] = parameter_value
                        if include_stderr:
                            df.loc[f'{parameter_name} (std)', col] = std_err_value
                        if include_t_test:
                            df.loc[f'{parameter_name} (t-test)', col] = t_test_value

    return df.fillna(''), configurations


def compile_results_in_directory(
    statistics: tuple[str, ...] = (
        'Number of estimated parameters',
        'Sample size',
        'Final log likelihood',
        'Akaike Information Criterion',
        'Bayesian Information Criterion',
    ),
    file_extension: str = 'yaml',
    variance_covariance_type: EstimateVarianceCovariance = EstimateVarianceCovariance.ROBUST,
    include_parameter_estimates: bool = True,
    include_stderr: bool = False,
    include_t_test: bool = True,
    formatted: bool = True,
    use_short_names: bool = False,
) -> tuple[pd.DataFrame, dict[str, str]] | None:
    """Compile estimation results found in the local directory into a
        common table. The results are supposed to be in a file with
        pickle extension.

    :param statistics: list of statistics to include in the summary
        table
    :param file_extension: extension of the files containing the estimation results.
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param include_parameter_estimates: if True, the parameter
        estimates are included.
    :param include_stderr: if True, the robust standard errors
         of the parameters are included.
    :param include_t_test: if True, the t-test
         of the parameters are included.
    :param formatted: if True, a formatted string in included in the
         table results. If False, the numerical values are stored. Use
         "True" if you need to print the results. Use "False" if you
         need to use them for further calculation.
    :param use_short_names: if True, short names, such as Model_1,
        Model_2, are used to identify the model. It is nicer on for the
        reporting.
    :return: pandas dataframe with the requested results, and a dictionary reporting the
        specification of each model

    """
    files = glob.glob(f'*.{file_extension}')
    if not files:
        logger.warning(f'No .{file_extension} file found in {os.getcwd()}')
        return None

    the_dict = {k: k for k in files}
    return compile_estimation_results(
        dict_of_results=the_dict,
        variance_covariance_type=variance_covariance_type,
        statistics=statistics,
        include_parameter_estimates=include_parameter_estimates,
        include_stderr=include_stderr,
        include_t_test=include_t_test,
        formatted=formatted,
        use_short_names=use_short_names,
    )



"""
Test the database module

:author: Michel Bierlaire
:data: Wed Apr 29 18:02:59 2020

"""

import unittest

import numpy as np
import pandas as pd

from biogeme.database import Database, sample_with_replacement
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Variable
from biogeme.segmentation import DiscreteSegmentationTuple
from test_data import (
    getData,
)


class TestDatabase(unittest.TestCase):
    def setUp(self):
        np.random.seed(90267)
        self.myData1 = getData(1)
        self.myPanelData = getData(3)
        self.Variable1 = Variable('Variable1')
        self.Variable2 = Variable('Variable2')
        self.Av1 = Variable('Av1')
        self.Av2 = Variable('Av2')
        self.Av3 = Variable('Av3')
        self.Choice = Variable('Choice')

    def test_empty_database(self):
        df = pd.DataFrame()
        with self.assertRaises(BiogemeError):
            _ = Database('name', df)

    def test_bad_values(self):
        with self.assertRaises(BiogemeError):
            df = getData(0)

    def test_check_segmentation(self):
        the_data = getData(1)
        correct_mapping = {1: 'Alt. 1', 2: 'Alt. 2', 3: 'Alt. 3'}
        correct_segmentation = DiscreteSegmentationTuple(
            variable='Choice', mapping=correct_mapping
        )
        # Successful check
        output = the_data.verify_segmentation(correct_segmentation)
        incorrect_mapping = {1: 'Alt. 1', 2: 'Alt. 2'}
        incorrect_segmentation = DiscreteSegmentationTuple(
            variable='Choice', mapping=incorrect_mapping
        )
        with self.assertRaises(BiogemeError):
            the_data.verify_segmentation(incorrect_segmentation)

        another_incorrect_mapping = {1: 'Alt. 1', 2: 'Alt. 2', 4: 'Does not exist'}
        another_incorrect_segmentation = DiscreteSegmentationTuple(
            variable='Choice', mapping=another_incorrect_mapping
        )
        with self.assertRaises(BiogemeError):
            the_data.verify_segmentation(another_incorrect_segmentation)

    def test_scale_column(self):
        self.myData1.scale_column('Variable1', 100)
        the_result = list(self.myData1.dataframe['Variable1'])
        expected_result = [100, 200, 300, 400, 500]
        self.assertListEqual(the_result, expected_result)

    def test_suggest_scaling(self):
        result = self.myData1.suggest_scaling()
        result.reset_index(inplace=True, drop=True)
        expected_result = pd.DataFrame(
            {
                'Column': ['Variable2'],
                'Scale': [0.01],
                'Largest': [50.0],
            }
        )
        pd.testing.assert_frame_equal(result, expected_result, check_index_type=False)
        result = self.myData1.suggest_scaling(columns=['Exclude', 'Variable2'])
        result.reset_index(inplace=True, drop=True)
        pd.testing.assert_frame_equal(result, expected_result, check_index_type=False)
        with self.assertRaises(BiogemeError):
            _ = self.myData1.suggest_scaling(columns=['wrong_name'])

    def test_addColumn(self):
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        expression = Variable2 * Variable1
        the_values = [10, 40, 90, 160, 250]
        self.myData1.add_column(column='NewVariable', values=the_values)
        the_list = self.myData1.dataframe['NewVariable'].tolist()
        self.assertListEqual(the_list, the_values)
        with self.assertRaises(ValueError):
            result = self.myData1.add_column(column='Variable1', values=the_values)
        database = getData(1)
        database._df = database.dataframe[0:0]
        with self.assertRaises(ValueError):
            result = database.add_column(column='NewVariable', values=the_values)

    def test_DefineVariable(self):
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        expression = Variable2 * Variable1
        result = self.myData1.define_variable('NewVariable', expression)
        theList = self.myData1.dataframe['NewVariable'].tolist()
        self.assertListEqual(theList, [10, 40, 90, 160, 250])
        with self.assertRaises(ValueError):
            result = self.myData1.define_variable('Variable1', expression)
        database = getData(1)
        database._df = database.dataframe[0:0]
        with self.assertRaises(BiogemeError):
            result = database.define_variable('NewVariable', expression)

    def test_remove(self):
        Exclude = Variable('Exclude')
        d = getData(1)
        d.remove(Exclude)
        rows, _ = d.dataframe.shape
        self.assertEqual(rows, 3)
        # This should not remove anything
        d.remove(0)
        rows, _ = d.dataframe.shape
        self.assertEqual(rows, 3)

    def test_sampleWithReplacement(self):
        res1 = sample_with_replacement(df=self.myData1.dataframe)
        res2 = sample_with_replacement(df=self.myData1.dataframe, size=12)
        dim1 = res1.shape
        dim2 = res2.shape
        self.assertTupleEqual(dim1, (5, 8))
        self.assertTupleEqual(dim2, (12, 8))

    def test_panel(self):
        # Data is not considered panel yet
        should_be_false = self.myPanelData.is_panel()
        self.myPanelData.panel('Person')
        should_be_true = self.myPanelData.is_panel()
        self.assertTrue(should_be_true)
        self.assertFalse(should_be_false)

    def test_print(self):
        result = str(self.myData1)[0:23]
        expected_result = 'biogeme database test_1'
        self.assertEqual(result, expected_result)
        self.myPanelData.panel('Person')
        result = str(self.myPanelData)[0:23]
        expected_result = 'biogeme database test_3'
        self.assertEqual(result, expected_result)

    def test_getNumberOfObservations(self):
        self.myPanelData.panel('Person')
        self.assertEqual(self.myData1.num_rows(), 5)
        self.assertEqual(self.myPanelData.num_rows(), 5)


class TestVerifySegmentation(unittest.TestCase):
    def setUp(self):
        self.my_data = getData(1)

    def test_verify_segmentation_valid(self):
        # Define a valid segmentation
        valid_segmentation = DiscreteSegmentationTuple(
            variable='Choice', mapping={1: 'Alt. 1', 2: 'Alt. 2', 3: 'Alt. 3'}
        )

        # Verify segmentation
        self.assertIsNone(self.my_data.verify_segmentation(valid_segmentation))

    def test_verify_segmentation_invalid_variable(self):
        # Define a segmentation with an invalid variable
        invalid_variable_segmentation = DiscreteSegmentationTuple(
            variable='invalid_variable', mapping={1: 'Alt. 1', 2: 'Alt. 2', 3: 'Alt. 3'}
        )

        # Verify segmentation and expect BiogemeError to be raised
        with self.assertRaises(BiogemeError):
            self.my_data.verify_segmentation(invalid_variable_segmentation)

    def test_verify_segmentation_missing_entries(self):
        # Define a segmentation with missing entries in the data
        missing_entries_segmentation = DiscreteSegmentationTuple(
            variable='Choice',
            mapping={
                1: 'Alt. 1',
                2: 'Alt. 2',
            },
        )

        # Verify segmentation and expect BiogemeError to be raised
        with self.assertRaises(BiogemeError) as cm:
            self.my_data.verify_segmentation(missing_entries_segmentation)
        self.assertIn('missing in the segmentation', str(cm.exception))

    def test_verify_segmentation_extra_entries(self):
        # Define a segmentation with extra entries not present in the data
        extra_entries_segmentation = DiscreteSegmentationTuple(
            variable='Choice',
            mapping={1: 'Alt. 1', 2: 'Alt. 2', 3: 'Alt. 3', 4: 'Alt. 4'},
        )
        # Verify segmentation and expect BiogemeError to be raised
        with self.assertRaises(BiogemeError) as cm:
            self.my_data.verify_segmentation(extra_entries_segmentation)
        self.assertIn('do not exist in the data', str(cm.exception))


class TestGenerateSegmentation(unittest.TestCase):
    def setUp(self):
        self.my_data = getData(1)

    def test_generate_segmentation_valid(self):
        # Define a valid variable and mapping
        variable = 'Choice'
        mapping = {1: 'Alt. 1', 2: 'Alt. 2', 3: 'Alt. 3'}

        # Generate segmentation
        segmentation = self.my_data.generate_segmentation(variable, mapping)

        expected_segmentation = DiscreteSegmentationTuple(
            variable='Choice', mapping=mapping
        )
        # Verify the generated segmentation
        self.assertIsInstance(segmentation, DiscreteSegmentationTuple)
        self.assertEqual(
            repr(segmentation.variable), repr(expected_segmentation.variable)
        )
        self.assertDictEqual(segmentation.mapping, expected_segmentation.mapping)

        segmentation = self.my_data.generate_segmentation(
            variable, mapping, reference='Alt. 2'
        )

        expected_segmentation = DiscreteSegmentationTuple(
            variable='Choice', mapping=mapping, reference='Alt. 2'
        )
        # Verify the generated segmentation
        self.assertIsInstance(segmentation, DiscreteSegmentationTuple)
        self.assertEqual(
            repr(segmentation.variable), repr(expected_segmentation.variable)
        )
        self.assertDictEqual(segmentation.mapping, expected_segmentation.mapping)
        self.assertEqual(segmentation.reference, expected_segmentation.reference)

        mapping = {1: 'Alt. 1', 2: 'Alt. 2'}
        expected_mapping = {1: 'Alt. 1', 2: 'Alt. 2', 3: 'Choice_3'}

        # Generate segmentation
        segmentation = self.my_data.generate_segmentation(variable, mapping)

        expected_segmentation = DiscreteSegmentationTuple(
            variable='Choice', mapping=expected_mapping
        )
        # Verify the generated segmentation
        self.assertIsInstance(segmentation, DiscreteSegmentationTuple)
        self.assertEqual(
            repr(segmentation.variable), repr(expected_segmentation.variable)
        )
        self.assertDictEqual(segmentation.mapping, expected_segmentation.mapping)

    def test_generate_segmentation_invalid_reference(self):
        variable = 'Choice'
        mapping = {1: 'Alt. 1', 2: 'Alt. 2', 3: 'Alt. 3'}

        # Generate segmentation
        with self.assertRaises(BiogemeError):
            _ = self.my_data.generate_segmentation(
                variable, mapping, reference='invalid_reference'
            )

    def test_generate_segmentation_invalid_variable(self):
        # Define an invalid variable
        invalid_variable = 'invalid_variable'
        mapping = {1: 'One', 2: 'Two'}

        # Generate segmentation and expect BiogemeError to be raised
        with self.assertRaises(BiogemeError):
            self.my_data.generate_segmentation(invalid_variable, mapping)

    def test_generate_segmentation_values_not_in_data(self):
        # Define a variable and mapping with values not present in the data
        variable = 'Choice'
        mapping = {1: 'One', 2: 'Two', 4: 'Four'}

        # Generate segmentation and expect BiogemeError to be raised
        with self.assertRaises(BiogemeError) as cm:
            self.my_data.generate_segmentation(variable, mapping)
        self.assertIn('do not exist in the data', str(cm.exception))


if __name__ == '__main__':
    unittest.main()



"""Tests for sampling utilities

Michel Bierlaire
Wed Mar 26 21:02:23 2025
"""

import unittest

import pandas as pd

from biogeme.database.sampling import (
    sample_with_replacement,
    sample_panel_with_replacement,
    split_validation_sets,
)
from biogeme.exceptions import BiogemeError


class TestSamplingFunctions(unittest.TestCase):

    def setUp(self):
        self.df = pd.DataFrame(
            {'id': [1, 1, 2, 2, 3, 3], 'value': [10, 20, 30, 40, 50, 60]}
        )

        self.individual_map = pd.DataFrame.from_dict(
            {1: [0, 1], 2: [2, 3], 3: [4, 5]}, orient='index'
        )

    def test_sample_with_replacement_default(self):
        result = sample_with_replacement(self.df)
        self.assertEqual(len(result), len(self.df))
        self.assertTrue(set(result.columns) == set(self.df.columns))

    def test_sample_with_replacement_custom_size(self):
        size = 10
        result = sample_with_replacement(self.df, size=size)
        self.assertEqual(len(result), size)

    def test_sample_panel_with_replacement_default(self):
        result = sample_panel_with_replacement(self.df, self.individual_map)
        self.assertTrue(len(result) % 2 == 0)  # Each individual contributes 2 rows

    def test_sample_panel_with_replacement_custom_size(self):
        result = sample_panel_with_replacement(self.df, self.individual_map, size=5)
        self.assertEqual(len(result) % 2, 0)
        self.assertEqual(len(result) // 2, 5)

    def test_sample_panel_with_replacement_invalid_map(self):
        with self.assertRaises(BiogemeError):
            sample_panel_with_replacement(self.df, None)

        with self.assertRaises(BiogemeError):
            sample_panel_with_replacement(self.df, pd.DataFrame())

    def test_split_validation_sets_no_grouping(self):
        result = split_validation_sets(self.df, slices=3)
        self.assertEqual(len(result), 3)
        for est, val in result:
            self.assertEqual(len(est) + len(val), len(self.df))

    def test_split_validation_sets_with_grouping(self):
        result = split_validation_sets(self.df, slices=2, group_column='id')
        self.assertEqual(len(result), 2)
        for est, val in result:
            self.assertEqual(len(est) + len(val), len(self.df))

    def test_split_validation_sets_invalid_slices(self):
        with self.assertRaises(BiogemeError):
            split_validation_sets(self.df, slices=1)

    def test_split_validation_sets_invalid_group_column(self):
        with self.assertRaises(BiogemeError):
            split_validation_sets(self.df, slices=2, group_column='not_a_column')


if __name__ == '__main__':
    unittest.main()



import unittest

import pandas as pd

from biogeme.database import Database, PanelDatabase
from biogeme.exceptions import BiogemeError


class TestFlattenDataFrame(unittest.TestCase):

    def test_contiguous_observations(self):
        df = pd.DataFrame(
            {
                'ID': [1, 1, 2, 2],
                'X': [10, 20, 30, 40],
            }
        )
        database = Database(name='test', dataframe=df)
        panel_database = PanelDatabase(database=database, panel_column='ID')
        result, largest_group = panel_database.flatten_database(missing_data=-999)
        self.assertEqual(result.shape[0], 2)
        self.assertIn('X__panel__01', result.columns)
        self.assertIn('X__panel__02', result.columns)
        self.assertTrue((result['relevant___panel__01'] == 1).all())
        self.assertTrue((result['relevant___panel__02'] == 1).all())
        self.assertEqual(largest_group, 2)

    def test_non_contiguous_observations(self):
        df = pd.DataFrame(
            {
                'ID': [2, 1, 2, 1],
                'X': [30, 10, 40, 20],
            }
        )
        database = Database(name='test', dataframe=df)
        panel_database = PanelDatabase(database=database, panel_column='ID')
        result, largest_group = panel_database.flatten_database(missing_data=-999)
        self.assertEqual(result.shape[0], 2)
        self.assertEqual(result.loc[result['ID'] == 1, 'X__panel__01'].values[0], 10)
        self.assertEqual(result.loc[result['ID'] == 1, 'X__panel__02'].values[0], 20)
        self.assertEqual(largest_group, 2)

    def test_single_observation_per_individual(self):
        df = pd.DataFrame(
            {
                'ID': [1, 2],
                'X': [100, 200],
            }
        )
        database = Database(name='test', dataframe=df)
        panel_database = PanelDatabase(database=database, panel_column='ID')
        result, largest_group = panel_database.flatten_database(missing_data=-999)
        self.assertTrue((result['relevant___panel__01'] == 1).all())
        self.assertEqual(largest_group, 1)

    def test_missing_grouping_column(self):
        df = pd.DataFrame(
            {
                'A': [1, 2],
                'X': [10, 20],
            }
        )
        with self.assertRaises(KeyError):
            database = Database(name='test', dataframe=df)
            panel_database = PanelDatabase(database=database, panel_column='ID')
            _ = panel_database.flatten_database(missing_data=-999)

    def test_missing_data_padding(self):
        df = pd.DataFrame(
            {
                'ID': [1, 1, 2],
                'X': [10, 20, 30],
            }
        )
        database = Database(name='test', dataframe=df)
        panel_database = PanelDatabase(database=database, panel_column='ID')
        result, largest_group = panel_database.flatten_database(missing_data=-1)
        self.assertEqual(result.loc[result['ID'] == 2, 'X__panel__02'].values[0], -1)
        self.assertEqual(
            result.loc[result['ID'] == 2, 'relevant___panel__02'].values[0], 0
        )
        self.assertEqual(largest_group, 2)

    def test_column_naming(self):
        df = pd.DataFrame(
            {
                'ID': [1, 1],
                'X': [5, 6],
                'Y': [7, 8],
            }
        )
        database = Database(name='test', dataframe=df)
        panel_database = PanelDatabase(database=database, panel_column='ID')
        result, largest_group = panel_database.flatten_database(missing_data=0)
        self.assertIn('X__panel__01', result.columns)
        self.assertIn('Y__panel__02', result.columns)
        self.assertEqual(result['X__panel__01'].values[0], 5)
        self.assertEqual(result['Y__panel__02'].values[0], 8)
        self.assertEqual(largest_group, 2)

    def test_empty_dataframe(self):
        df = pd.DataFrame(columns=['ID', 'X'])
        with self.assertRaises(BiogemeError):
            database = Database(name='test', dataframe=df)


if __name__ == '__main__':
    unittest.main()



import unittest

import biogeme.biogeme as bio
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.expressions import Beta, Draws, MonteCarlo, PanelLikelihoodTrajectory, log

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)

database.panel('ID')

ASC_CAR = Beta('ASC_CAR', -0.165109, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', -0.756938, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', -0.756938, None, None, 0)
B_COST = Beta('B_COST', -0.756938, None, None, 0)

SIGMA_CAR = Beta('SIGMA_CAR', 0.262894, None, None, 0)
SIGMA_SM = Beta('SIGMA_SM', 0, None, None, 1)
SIGMA_TRAIN = Beta('SIGMA_TRAIN', -0.756938, None, None, 0)


# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
EC_CAR = SIGMA_CAR * Draws('EC_CAR', 'NORMAL')
EC_SM = SIGMA_SM * Draws('EC_SM', 'NORMAL')
EC_TRAIN = SIGMA_TRAIN * Draws('EC_TRAIN', 'NORMAL')


SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED + EC_TRAIN
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED + EC_SM
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED + EC_CAR

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

obs_prob = models.logit(V, av, CHOICE)
cond_prob_indiv = PanelLikelihoodTrajectory(obs_prob)
log_prob = log(MonteCarlo(cond_prob_indiv))


class test_13(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(
            database,
            log_prob,
            number_of_draws=100,
            seed=111,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -3877.4693151936744, 2)


if __name__ == '__main__':
    unittest.main()



"""

Calculation of market shares
============================

We use an estimated model to calculate market shares.

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:03:52 2023

"""
import sys
from biogeme import models
import biogeme.biogeme as bio
import biogeme.exceptions as excep
import biogeme.results as res
from optima_data import database, normalized_weight
from scenarios import scenario

# %%
# Obtain the specification for the default scenario
V, nests, _, _ = scenario()

# %%
# Obtain the expression for the choice probability of each alternative.
prob_PT = models.nested(V, None, nests, 0)
prob_CAR = models.nested(V, None, nests, 1)
prob_SM = models.nested(V, None, nests, 2)

# %%
# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='saved_results/b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02simulation.py '
        'in order to generate the '
        'file b02estimation.pickle.'
    )

# %%
# We now simulate the choice probabilities and the weight
simulate = {
    'weight': normalized_weight,
    'Prob. PT': prob_PT,
    'Prob. car': prob_CAR,
    'Prob. SM': prob_SM,
}

the_biogeme = bio.BIOGEME(database, simulate)
simulated_values = the_biogeme.simulate(results.getBetaValues())

# %%
# We also calculate confidence intervals for the calculated quantities,
betas = the_biogeme.free_beta_names()
b = results.getBetasForSensitivityAnalysis(betas)
left, right = the_biogeme.confidenceIntervals(b, 0.9)

# %%
# Market shares are calculated using the weighted mean of the
# individual probabilities.

# %%
# Alternative car
simulated_values['Weighted prob. car'] = (
    simulated_values['weight'] * simulated_values['Prob. car']
)
left['Weighted prob. car'] = left['weight'] * left['Prob. car']
right['Weighted prob. car'] = right['weight'] * right['Prob. car']

marketShare_car = simulated_values['Weighted prob. car'].mean()
marketShare_car_left = left['Weighted prob. car'].mean()
marketShare_car_right = right['Weighted prob. car'].mean()

# %%
# Alternative public transportation
simulated_values['Weighted prob. PT'] = (
    simulated_values['weight'] * simulated_values['Prob. PT']
)
left['Weighted prob. PT'] = left['weight'] * left['Prob. PT']
right['Weighted prob. PT'] = right['weight'] * right['Prob. PT']

marketShare_PT = simulated_values['Weighted prob. PT'].mean()
marketShare_PT_left = left['Weighted prob. PT'].mean()
marketShare_PT_right = right['Weighted prob. PT'].mean()

# %%
# Alternative slow modes
simulated_values['Weighted prob. SM'] = (
    simulated_values['weight'] * simulated_values['Prob. SM']
)
left['Weighted prob. SM'] = left['weight'] * left['Prob. SM']
right['Weighted prob. SM'] = right['weight'] * right['Prob. SM']

marketShare_SM = simulated_values['Weighted prob. SM'].mean()
marketShare_SM_left = left['Weighted prob. SM'].mean()
marketShare_SM_right = right['Weighted prob. SM'].mean()

# %%
# Reporting.

# %%
# Car.
print(
    f'Market share for car: {100*marketShare_car:.1f}% '
    f'[{100*marketShare_car_left:.1f}%, '
    f'{100*marketShare_car_right:.1f}%]'
)

# %%
# Public transportation.
print(
    f'Market share for PT:  {100*marketShare_PT:.1f}% '
    f'[{100*marketShare_PT_left:.1f}%, '
    f'{100*marketShare_PT_right:.1f}%]'
)

# %%
# Slow modes.
print(
    f'Market share for SM:   {100*marketShare_SM:.1f}% '
    f'[{100*marketShare_SM_left:.1f}%, '
    f'{100*marketShare_SM_right:.1f}%]'
)



""" Implements piecewise linear formulation

:author: Michel Bierlaire
:date: Wed Oct 25 09:43:41 2023
"""

import logging

from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Expression,
    Beta,
    Variable,
    BinaryMin,
    BinaryMax,
    Numeric,
    MultipleSum,
    ExpressionOrNumeric,
)

logger = logging.getLogger(__name__)


def piecewise_variables(
    variable: Expression | str, thresholds: list[float]
) -> list[Expression]:
    """Generate the variables to include in a piecewise linear specification.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\\infty`,respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:

    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if }
              t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  &
              \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} =
              \\max(0, \\min(t-a, b))

    :param variable: variable for which we need the piecewise linear
       transform. The expression itself or the name of the variable
       can be given.

    :param thresholds: list of thresholds

    :return: list of variables to for the piecewise linear specification.

    :raise BiogemeError: if the thresholds are not defined properly,
        as only the first and the last thresholds can be set
        to None.

    .. seealso:: :meth:`piecewiseFormula`

    """
    eye = len(thresholds)
    if eye == 0:
        error_msg = 'No threshold has been provided.'
        raise BiogemeError(error_msg)
    if all(t is None for t in thresholds):
        error_msg = (
            'All thresholds for the piecewise linear specification are set to None.'
        )
        raise BiogemeError(error_msg)
    if None in thresholds[1:-1]:
        error_msg = (
            'For piecewise linear specification, only the first and '
            'the last thresholds can be None'
        )
        raise BiogemeError(error_msg)

    # If the name of the variable is given, we transform it into an expression.
    if isinstance(variable, str):
        variable = Variable(variable)
    if not isinstance(variable, Variable):
        error_msg = f'Expression of type Variable expected, not {type(variable)}'
        raise BiogemeError(error_msg)

    # First variable
    if thresholds[0] is None:
        results = [BinaryMin(variable, thresholds[1])]
    else:
        b = thresholds[1] - thresholds[0]
        results = [BinaryMax(Numeric(0), BinaryMin(variable - thresholds[0], b))]

    for i in range(1, eye - 2):
        b = thresholds[i + 1] - thresholds[i]
        results += [BinaryMax(Numeric(0), BinaryMin(variable - thresholds[i], b))]

    # Last variable
    if thresholds[-1] is None:
        results += [BinaryMax(0, variable - thresholds[-2])]
    else:
        b = thresholds[-1] - thresholds[-2]
        results += [BinaryMax(Numeric(0), BinaryMin(variable - thresholds[-2], b))]
    return results


def piecewise_formula(
    variable: str | Variable,
    thresholds: list[float],
    betas: list[ExpressionOrNumeric] | None = None,
) -> Expression:
    """Generate the formula for a piecewise linear specification.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\\infty`, respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:

    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if }
              t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  &
              \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} =
              \\max(0, \\min(t-a, b))

    New variables and new parameters are automatically created to
    obtain the specification

    .. math:: \\sum_{i=1}^{K-1} \\beta_i x_{Ti}

    :param variable: name of the variable for which we need the
        piecewise linear transform.

    :param thresholds: list of thresholds

    :param betas: list of Beta parameters to be used in the
        specification.  The number of entries should be the number of
        thresholds, minus one. If None, for each interval, the
        parameter Beta('beta_VAR_interval',0, None, None, 0) is used,
        where var is the name of the variable. Default: none.

    :return: expression of  the piecewise linear specification.

    :raise BiogemeError: if the thresholds are not defined properly,
        which means that only the first and the last threshold can be set
        to None.

    :raise BiogemeError: if the length of list ``initialexpr.Betas`` is
        not equal to the length of ``thresholds`` minus one.

    .. seealso:: :meth:`piecewiseVariables`

    """
    if isinstance(variable, Variable):
        the_variable = variable
        the_name = variable.name
    elif isinstance(variable, str):
        the_name = variable
        the_variable = Variable(f'{variable}')
    else:
        error_msg = (
            'The first argument of piecewiseFormula must be the '
            'name of a variable, or the variable itself..'
        )
        raise BiogemeError(error_msg)

    eye = len(thresholds)
    if all(t is None for t in thresholds):
        error_msg = (
            'All thresholds for the piecewise linear specification are set to None.'
        )
        raise BiogemeError(error_msg)
    if None in thresholds[1:-1]:
        error_msg = (
            'For piecewise linear specification, only the first and '
            'the last thresholds can be None'
        )
        raise BiogemeError(error_msg)
    if betas is not None:
        if len(betas) != eye - 1:
            error_msg = (
                f'As there are {eye} thresholds, a total of {eye-1} '
                f'Beta parameters are needed, and not {len(betas)}.'
            )
            raise BiogemeError(error_msg)

    the_vars = piecewise_variables(the_variable, thresholds)
    if betas is None:
        betas = []
        for i, a_threshold in enumerate(thresholds[:-1]):
            next_threshold = thresholds[i + 1]
            a_name = 'minus_inf' if a_threshold is None else f'{a_threshold}'
            next_name = 'inf' if next_threshold is None else f'{next_threshold}'
            betas.append(
                Beta(f'beta_{the_name}_{a_name}_{next_name}', 0, None, None, 0)
            )

    terms = [beta * the_vars[i] for i, beta in enumerate(betas)]

    return MultipleSum(terms)


def piecewise_as_variable(
    variable: str | Variable, thresholds: list[float], betas: list[Beta] | None = None
) -> Expression:
    """Generate the formula for a piecewise linear specification, seen
    as a transformed variable.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\\infty`, respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:

    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if }
              t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  &
              \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} =
              \\max(0, \\min(t-a, b))

    The specification this is returned is

    .. math:: x_{T1} + \\sum_{i=2}^{K-1} \beta_i x_{Ti}

    :param variable: name of the variable for which we need the
        piecewise linear transform.
    :type variable: string

    :param thresholds: list of thresholds
    :type thresholds: list(float)

    :param betas: list of Beta parameters to be used in the
        specification.  The number of entries should be the number of
        thresholds, minus two. If None, for each interval, the
        parameter Beta('beta_VAR_interval',0, None, None, 0) is used,
        where var is the name of the variable. Default: none.
    :type betas:
        list(biogeme.expresssions.Beta)

    :return: expression of  the piecewise linear specification.
    :rtype: biogeme.expressions.expr.Expression

    :raise BiogemeError: if the thresholds are not defined properly,
        which means that only the first and the last threshold can be set
        to None.

    :raise BiogemeError: if the length of list ``initialexpr.Betas`` is
        not equal to the length of ``thresholds`` minus one.

    .. see also:: :meth:`piecewiseVariables`

    """
    if isinstance(variable, Variable):
        the_variable = variable
        the_name = variable.name
    elif isinstance(variable, str):
        the_name = variable
        the_variable = Variable(f'{variable}')
    else:
        error_msg = (
            'The first argument of piecewiseFormula must be the '
            'name of a variable, or the variable itself..'
        )
        raise BiogemeError(error_msg)

    eye = len(thresholds)
    if all(t is None for t in thresholds):
        error_msg = (
            'All thresholds for the piecewise linear specification are set to None.'
        )
        raise BiogemeError(error_msg)
    if None in thresholds[1:-1]:
        error_msg = (
            'For piecewise linear specification, only the first and '
            'the last thresholds can be None'
        )
        raise BiogemeError(error_msg)
    if betas is not None:
        if len(betas) != eye - 2:
            error_msg = (
                f'As there are {eye} thresholds, a total of {eye-2} '
                f'Beta parameters are needed, and not {len(betas)}.'
            )
            raise BiogemeError(error_msg)

    theVars = piecewise_variables(the_variable, thresholds)
    if betas is None:
        betas = []
        for i, a_threshold in enumerate(thresholds[1:-1]):
            next_threshold = thresholds[i + 2]
            a_name = 'minus_inf' if a_threshold is None else f'{a_threshold}'
            next_name = 'inf' if next_threshold is None else f'{next_threshold}'
            betas.append(
                Beta(f'beta_{the_name}_{a_name}_{next_name}', 0, None, None, 0)
            )

    terms = [beta * theVars[i] for i, beta in enumerate(betas)]

    return theVars[0] + MultipleSum(terms)


def piecewise_function(x: float, thresholds: list[float], betas: list[float]) -> float:
    """Plot a piecewise linear specification.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\\infty`, respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:

    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if }
              t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  &
              \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} =
              \\max(0, \\min(t-a, b))

    :param x: value at which the piecewise specification must be avaluated
    :type x: float

    :param thresholds: list of thresholds
    :type thresholds: list(float)

    :param betas: list of the Beta parameters.  The number of entries
                         should be the number of thresholds, plus
                         one.
    :type betas: list(float)

    :return: value of the numpy function
    :rtype: float

    :raise BiogemeError: if the thresholds are not defined properly,
        which means that only the first and the last threshold can be set
        to None.


    """
    eye = len(thresholds)
    if all(t is None for t in thresholds):
        error_msg = (
            'All thresholds for the piecewise linear specification are set to None.'
        )
        raise BiogemeError(error_msg)
    if None in thresholds[1:-1]:
        error_msg = (
            'For piecewise linear specification, only the first and '
            'the last thresholds can be None'
        )
        raise BiogemeError(error_msg)
    if len(betas) != eye - 1:
        error_msg = (
            f'As there are {eye} thresholds, a total of {eye-1} values '
            f'are needed to initialize the parameters. But '
            f'{len(betas)} are provided'
        )
        raise BiogemeError(error_msg)

    # If the first threshold is not -infinity, we need to check if
    # x is beyond it.
    if thresholds[0] is not None:
        if x < thresholds[0]:
            return 0
    rest = x
    total = 0
    for i, v in enumerate(betas):
        if thresholds[i + 1] is None:
            total += v * rest
            return total
        if x < thresholds[i + 1]:
            total += v * rest
            return total
        total += v * (
            thresholds[i + 1] - (0 if thresholds[i] is None else thresholds[i])
        )
        rest = x - thresholds[i + 1]
    return total



"""Sets the type of floating point used by Biogeme

Michel Bierlaire
Mon Mar 31 09:57:37 2025
"""

import logging
import os

import jax.numpy as jnp
import numpy as np
from jax import config

from biogeme.exceptions import BiogemeError

logger = logging.getLogger(__name__)

raw_value = os.getenv('BIOGEME_FLOAT_TYPE', '64')
if raw_value not in {'32', '64'}:
    logger.warning(
        f'Invalid float type: {raw_value}. Valid values: "32" or "64". Update the environment variable BIOGEME_FLOAT_TYPE. "64" is used by default'
    )
    raw_value = '64'
FLOAT_TYPE = int(raw_value)

if FLOAT_TYPE == 64:
    config.update("jax_enable_x64", True)

if FLOAT_TYPE != 64 and FLOAT_TYPE != 32:
    raise BiogemeError('FLOAT_TYPE must be 32 or 64')

FLOAT = 'float64' if FLOAT_TYPE == 64 else 'float32'
NUMPY_FLOAT = np.float64 if FLOAT_TYPE == 64 else np.float32
PANDAS_FLOAT = FLOAT
JAX_FLOAT = jnp.dtype(jnp.float32) if FLOAT_TYPE == 32 else jnp.dtype(jnp.float64)
EPSILON = jnp.finfo(JAX_FLOAT).eps
SQRT_EPS = jnp.sqrt(EPSILON)
LOG_CLIP_MIN = SQRT_EPS
MAX_EXP_ARG = jnp.log(jnp.finfo(JAX_FLOAT).max)
MIN_EXP_ARG = jnp.log(jnp.finfo(JAX_FLOAT).tiny)
MOST_NEGATIVE = jnp.finfo(JAX_FLOAT).min



from __future__ import annotations

import pandas as pd

from biogeme.audit_tuple import AuditTuple, display_messages, merge_audit_tuples
from biogeme.constants import LOG_LIKE, WEIGHT
from biogeme.database import Database, PanelDatabase, audit_dataframe
from biogeme.default_parameters import MISSING_VALUE
from biogeme.draws import DrawsManagement, RandomNumberGeneratorTuple
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression, audit_expression
from biogeme.expressions.base_expressions import LogitTuple
from biogeme.expressions_registry import ExpressionRegistry
from biogeme.function_output import FunctionOutput, NamedFunctionOutput
from biogeme.model_elements.audit import audit_chosen_alternative, audit_variables


class ModelElements:
    """
    Container for all key components required to define and estimate a model.

    :param expressions: Dict of expressions to be evaluated.
    """

    def __init__(
        self,
        expressions: dict[str, Expression],
        database: Database | None,
        number_of_draws: int | None = None,
        draws_management: DrawsManagement | None = None,
        user_defined_draws: dict[str:RandomNumberGeneratorTuple] | None = None,
        expressions_registry: ExpressionRegistry = None,
    ):
        if database is None:
            database = Database.dummy_database()
        self.panel_prepared: bool = False
        if database.panel_column is None:
            self._is_panel = False
            self._panel_database: PanelDatabase | None = None
            self._database: Database = database
        else:
            self._is_panel: bool = True
            self._panel_database = PanelDatabase(
                database=database, panel_column=database.panel_column
            )
            self._database: Database = database

        self.expressions: dict[str, Expression] = expressions
        self._flat_database: Database | None = None
        self.number_of_draws: int = number_of_draws
        self.draws_management = draws_management
        self.user_defined_draws = user_defined_draws
        self.expressions_registry = expressions_registry
        # Validate inputs: only one must be provided
        if self.number_of_draws and self.draws_management:
            raise ValueError(
                "One of 'number_of_draws' or 'draws_management' must be provided (or none of them)."
            )

        if self.expressions_registry is None:
            self.expressions_registry = ExpressionRegistry(
                self.expressions.values(), self.database
            )

        if self._database is not None:
            self.database.register_listener(self.on_database_update)

            if self.draws_management is None:
                self.draws_management = DrawsManagement(
                    sample_size=self.sample_size,
                    number_of_draws=self.number_of_draws,
                    user_generators=user_defined_draws,
                )
                if self.expressions_registry.requires_draws:
                    self.draws_management.generate_draws(
                        draw_types=self.expressions_registry.draw_types(),
                        variable_names=self.expressions_registry.draws_names,
                    )
            else:
                # Check consistency of sized
                if self.draws_management.sample_size != self.sample_size:
                    error_msg = f'Inconsistent sizes: database[{self.sample_size}] and draws [{self.draws_management.sample_size}]'
                    raise BiogemeError(error_msg)

        if self._is_panel:
            self._prepare_for_panel()
            self.panel_prepared: bool = True
        display_messages(self.audit())

    def is_panel(self) -> bool:
        return self._is_panel

    @property
    def database(self) -> Database:
        if not self.is_panel():
            return self._database
        if not self.panel_prepared:
            self._prepare_for_panel()
        return self._flat_database

    @classmethod
    def from_expression_and_weight(
        cls,
        log_like: Expression,
        database: Database,
        weight: Expression | None = None,
        number_of_draws: int = 0,
        draws_management: DrawsManagement | None = None,
        user_defined_draws: dict[str:RandomNumberGeneratorTuple] | None = None,
    ) -> ModelElements:
        """
        Alternative constructor for two expressions.

        :param log_like: Expression for the log-likelihood.
        :param weight: Expression for the weight.
        :param database: The database containing data.
        :param number_of_draws: Number of Monte Carlo draws.
        :param draws_management: Optional object managing the draws.
        """
        expressions = (
            {LOG_LIKE: log_like}
            if weight is None
            else {LOG_LIKE: log_like, WEIGHT: weight}
        )
        return cls(
            expressions=expressions,
            database=database,
            number_of_draws=number_of_draws,
            draws_management=draws_management,
            user_defined_draws=user_defined_draws,
        )

    def on_database_update(self, updated_index: pd.Index):
        """Update the draws object to remain consistent with the new database"""
        self.draws_management.remove_rows(updated_index)
        self.panel_prepared = False

    def _prepare_for_panel(self) -> None:

        flat_dataframe, maximum_number_of_observations_per_individual = (
            self._panel_database.flatten_database(missing_data=MISSING_VALUE)
        )
        self._flat_database = Database(
            name=f'flat {self._database.name}', dataframe=flat_dataframe
        )
        for expression in self.expressions.values():
            expression.set_maximum_number_of_observations_per_individual(
                max_number=maximum_number_of_observations_per_individual
            )
        self.expressions_registry = ExpressionRegistry(
            self.expressions.values(), self._flat_database
        )

    @property
    def sample_size(self) -> int:
        if self._database is None:
            return 0
        return self.database.num_rows()

    @property
    def number_of_observations(self) -> int:
        if self._database is None:
            return 0
        return self._database.num_rows()

    @property
    def loglikelihood(self) -> Expression | None:
        return self.expressions.get(LOG_LIKE)

    @property
    def weight(self) -> Expression | None:
        return self.expressions.get(WEIGHT)

    @property
    def formula_names(self) -> list[str]:
        return list(self.expressions.keys())

    def generate_named_output(
        self, function_output: FunctionOutput
    ) -> NamedFunctionOutput:
        """Assigns parameter name to the entries of the gradient and the hessian"""
        return NamedFunctionOutput(
            function_output=function_output,
            mapping=self.expressions_registry.free_betas_indices,
        )

    def audit(self) -> AuditTuple:
        """Audit the model elements"""

        # First, we audit the expressions.
        expression_audits = [
            audit_expression(expr) for expr in self.expressions.values()
        ]

        # Second, we audit the database
        database_audits = (
            [audit_dataframe(data=self.database.dataframe)]
            if self._database is not None
            else []
        )

        # Then, we check the variables

        variables_audits = [
            audit_variables(expression=expr, database=self.database)
            for expr in self.expressions.values()
        ]

        # Finally, we verify the logit formula, if any.
        logit_audits = []
        if self.loglikelihood is not None:
            logits_to_check: list[LogitTuple] = self.loglikelihood.logit_choice_avail()
            logit_audits = (
                [
                    audit_chosen_alternative(
                        choice=logit_to_check.choice,
                        availability=logit_to_check.availabilities,
                        database=self.database,
                    )
                    for logit_to_check in logits_to_check
                ]
                if self.loglikelihood is not None
                else []
            )

        return merge_audit_tuples(
            expression_audits + database_audits + variables_audits + logit_audits
        )



"""

Simulation of a cross-nested logit model
========================================

Illustration of the application of an estimated model.

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:08:29 2023
"""

import sys
import biogeme.biogeme as bio
from biogeme import models
import biogeme.results as res
from biogeme.expressions import Beta, Derive
import biogeme.exceptions as excep
from biogeme.nests import OneNestForCrossNestedLogit, NestsForCrossNestedLogit

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_TT,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_TT,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_TT,
    CAR_CO_SCALED,
)

# %%
# Simulation must be done with estimated value of the
# parameters. We set them to zero here , and read the values from the
# file created after estimation.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Nest parameters.
MU_EXISTING = Beta('MU_EXISTING', 1, 1, None, 0)
MU_PUBLIC = Beta('MU_PUBLIC', 1, 1, None, 0)

# %%
# Nest membership parameters.
ALPHA_EXISTING = Beta('ALPHA_EXISTING', 0.5, 0, 1, 0)
ALPHA_PUBLIC = 1 - ALPHA_EXISTING

# %%
# Definition of the utility functions. Note that we need to have
# explicitly the unscaled variables if we want to calculate the
# elasticities.
V1 = ASC_TRAIN + B_TIME * TRAIN_TT / 100 + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT / 100 + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT / 100 + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of nests.

nest_existing = OneNestForCrossNestedLogit(
    nest_param=MU_EXISTING,
    dict_of_alpha={1: ALPHA_EXISTING, 2: 0.0, 3: 1.0},
    name='existing',
)

nest_public = OneNestForCrossNestedLogit(
    nest_param=MU_PUBLIC, dict_of_alpha={1: ALPHA_PUBLIC, 2: 1.0, 3: 0.0}, name='public'
)

nests = NestsForCrossNestedLogit(
    choice_set=[1, 2, 3], tuple_of_nests=(nest_existing, nest_public)
)

# %%
# Read the estimation results from the pickle file.
try:
    results = res.bioResults(pickleFile='saved_results/b11cnl.pickle')
except excep.BiogemeError:
    print('Run first the script b11cnl.py in order to generate the file 11cnl.pickle.')
    sys.exit()

# %%
print('Estimation results: ', results.getEstimatedParameters())

# %%
print('Calculating correlation matrix. It may generate numerical warnings from scipy.')
corr = nests.correlation(
    parameters=results.getBetaValues(),
    alternatives_names={1: 'Train', 2: 'Swissmetro', 3: 'Car'},
)
corr

# %%
# The choice model is a cross-nested logit, with availability conditions.
prob1 = models.cnl_avail(V, av, nests, 1)
prob2 = models.cnl_avail(V, av, nests, 2)
prob3 = models.cnl_avail(V, av, nests, 3)

# %%
# We calculate elasticities. It is important that the variables
# explicitly appear as such in the model. If not, the derivative will
# be zero, as well as the elasticities.
genelas1 = Derive(prob1, 'TRAIN_TT') * TRAIN_TT / prob1
genelas2 = Derive(prob2, 'SM_TT') * SM_TT / prob2
genelas3 = Derive(prob3, 'CAR_TT') * CAR_TT / prob3

# %%
# We report the probability of each alternative and the elasticities.
simulate = {
    'Prob. train': prob1,
    'Prob. Swissmetro': prob2,
    'Prob. car': prob3,
    'Elas. 1': genelas1,
    'Elas. 2': genelas2,
    'Elas. 3': genelas3,
}

# %%
# Create the Biogeme object.
biosim = bio.BIOGEME(database, simulate)
biosim.modelName = 'b11cnl_simul'

# %%
# Perform the simulation.
simresults = biosim.simulate(results.getBetaValues())

# %%
print('Simulation results')
simresults

# %%
print(f'Aggregate share of train: {100*simresults["Prob. train"].mean():.1f}%')



"""Tests for comparison expressions

Michel Bierlaire
Wed Mar 26 13:25:27 2025
"""

import unittest
from biogeme.expressions.comparison_expressions import (
    Equal,
    NotEqual,
    LessOrEqual,
    GreaterOrEqual,
    Less,
    Greater,
)
from biogeme.expressions import Numeric


class TestComparisonExpressions(unittest.TestCase):
    def setUp(self):
        self.a = Numeric(3)
        self.b = Numeric(5)
        self.c = Numeric(3)

    def test_equal(self):
        expr = Equal(self.a, self.c)
        self.assertEqual(expr.get_value(), 1.0)
        self.assertEqual(str(expr), '(`3.0` == `3.0`)')
        self.assertEqual(repr(expr), 'Equal(<Numeric value=3.0>, <Numeric value=3.0>)')

        expr_false = Equal(self.a, self.b)
        self.assertEqual(expr_false.get_value(), 0.0)

    def test_not_equal(self):
        expr = NotEqual(self.a, self.b)
        self.assertEqual(expr.get_value(), 1.0)
        self.assertEqual(str(expr), '(`3.0` != `5.0`)')
        self.assertEqual(
            repr(expr), "NotEqual(<Numeric value=3.0>, <Numeric value=5.0>)"
        )

        expr_false = NotEqual(self.a, self.c)
        self.assertEqual(expr_false.get_value(), 0.0)

    def test_less_or_equal(self):
        expr = LessOrEqual(self.a, self.b)
        self.assertEqual(expr.get_value(), 1.0)
        self.assertEqual(str(expr), '(`3.0` <= `5.0`)')
        self.assertEqual(
            repr(expr), "LessOrEqual(<Numeric value=3.0>, <Numeric value=5.0>)"
        )

        expr_equal = LessOrEqual(self.a, self.c)
        self.assertEqual(expr_equal.get_value(), 1.0)

        expr_false = LessOrEqual(self.b, self.a)
        self.assertEqual(expr_false.get_value(), 0.0)

    def test_greater_or_equal(self):
        expr = GreaterOrEqual(self.b, self.a)
        self.assertEqual(expr.get_value(), 1.0)
        self.assertEqual(str(expr), '(`5.0` >= `3.0`)')
        self.assertEqual(
            repr(expr), "GreaterOrEqual(<Numeric value=5.0>, <Numeric value=3.0>)"
        )

        expr_equal = GreaterOrEqual(self.a, self.c)
        self.assertEqual(expr_equal.get_value(), 1.0)

        expr_false = GreaterOrEqual(self.a, self.b)
        self.assertEqual(expr_false.get_value(), 0.0)

    def test_less(self):
        expr = Less(self.a, self.b)
        self.assertEqual(expr.get_value(), 1.0)
        self.assertEqual(str(expr), '(`3.0` < `5.0`)')
        self.assertEqual(repr(expr), "Less(<Numeric value=3.0>, <Numeric value=5.0>)")

        expr_false = Less(self.b, self.a)
        self.assertEqual(expr_false.get_value(), 0.0)

    def test_greater(self):
        expr = Greater(self.b, self.a)
        self.assertEqual(expr.get_value(), 1.0)
        self.assertEqual(str(expr), '(`5.0` > `3.0`)')
        self.assertEqual(
            repr(expr), "Greater(<Numeric value=5.0>, <Numeric value=3.0>)"
        )

        expr_false = Greater(self.a, self.b)
        self.assertEqual(expr_false.get_value(), 0.0)


if __name__ == "__main__":
    unittest.main()



# from .validation import validate_expression_type
from .add_prefix_suffix import add_prefix_suffix_to_all_variables
from .audit import audit_expression
from .base_expressions import (
    Expression,
    ExpressionOrNumeric,
)
from .belongs_to import BelongsTo
from .beta_parameters import Beta
from .binary_max import BinaryMax
from .binary_min import BinaryMin
from .collectors import (
    ExpressionCollector,
    collect_init_values,
    list_of_all_betas_in_expression,
    list_of_draws_in_expression,
    list_of_fixed_betas_in_expression,
    list_of_free_betas_in_expression,
    list_of_random_variables_in_expression,
    list_of_variables_in_expression,
)
from .conditional_sum import ConditionalSum, ConditionalTermTuple
from .convert import (
    expression_to_value,
    get_dict_expressions,
    get_dict_values,
    validate_and_convert,
)
from .cos import cos
from .deprecated import (
    Integrate,
    bioDraws,
    bioLinearUtility,
    bioMax,
    bioMin,
    bioMultSum,
    bioNormalCdf,
)
from .derive import Derive
from .draws import Draws
from .elem import Elem
from .elementary_expressions import (
    Elementary,
    get_free_beta_values,
)
from .elementary_types import TypeOfElementaryExpression
from .exp import exp
from .integrate import IntegrateNormal
from .jax_utils import build_vectorized_function
from .linear_utility import LinearTermTuple, LinearUtility
from .log import log
from .logit_expressions import LogLogit
from .logzero import logzero
from .montecarlo import MonteCarlo
from .multiple_expressions import (
    CatalogItem,
    MultipleExpression,
    NamedExpression,
    SELECTION_SEPARATOR,
    SEPARATOR,
)
from .multiple_product import MultipleProduct
from .multiple_sum import MultipleSum
from .named_expression import named_function_output
from .normalcdf import NormalCdf
from .numeric_expressions import Numeric
from .numeric_tools import is_numeric
from .panel_likelihood_trajectory import PanelLikelihoodTrajectory
from .random_variable import RandomVariable
from .rename_variables import rename_all_variables
from .sin import sin
from .variable import Variable
from .visitor import ExpressionVisitor



""" Interface for the use of the loggers in Biogeme

:author: Michel Bierlaire
:date: Thu Mar 23 17:53:22 2023

"""

import logging
from contextlib import contextmanager

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

loggers = (
    logging.getLogger('biogeme'),
    logging.getLogger('biogeme_optimization'),
)


def get_screen_logger(level: int = WARNING) -> logging.Logger:
    """Obtain a screen logger

    :param level: level of verbosity of the logger
    :type level: int
    """
    for logger in loggers:
        # This has to be set to the lower level: DEBUG so that it does not
        # supersed the handler
        logger.setLevel(DEBUG)
        formatter_debug = logging.Formatter(
            '[%(levelname)s] %(asctime)s %(message)s <%(filename)s:%(lineno)d>'
        )
        formatter_normal = logging.Formatter('%(message)s ')
        formatter = formatter_debug if level == DEBUG else formatter_normal
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(level)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
    return loggers[0]


def get_file_logger(filename: str, level: int = WARNING) -> logging.Logger:
    """Obtain a file logger

    :param filename: name of the file. Extension .log recommended.
    :type filename: str

    :param level: level of verbosity of the logger
    :type level: int
    """
    for logger in loggers:
        # This has to be set to the lower level: DEBUG so that it does not
        # supersed the handler
        logger.setLevel(DEBUG)
        formatter = logging.Formatter(
            '[%(levelname)s] %(asctime)s %(message)s <%(filename)s:%(lineno)d>'
        )
        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    return loggers[0]


@contextmanager
def suppress_logs(level=logging.WARNING):
    """Temporarily suppress logs below a certain level."""

    old_levels = [logger.getEffectiveLevel() for logger in loggers]

    # Apply a filter
    class MaxLevelFilter(logging.Filter):
        def filter(self, record):
            return record.levelno >= level

    filters = []
    for logger in loggers:
        for handler in logger.handlers:
            filt = MaxLevelFilter()
            handler.addFilter(filt)
            filters.append((handler, filt))

    try:
        yield
    finally:
        for handler, filt in filters:
            handler.removeFilter(filt)
        for logger, old_level in zip(loggers, old_levels):
            logger.setLevel(old_level)



import unittest

from biogeme.partition import Partition


class TestPartition(unittest.TestCase):
    def test_valid_partition(self):
        """Test a valid partition."""
        segments = [{1, 2}, {3, 4, 5}]
        full_set = {1, 2, 3, 4, 5}
        partition = Partition(segments, full_set)
        self.assertEqual(list(partition), segments)

    def test_invalid_segment_type(self):
        """Test for a segment that is not of type 'set'."""
        segments = [[1, 2], {3, 4, 5}]
        full_set = {1, 2, 3, 4, 5}
        with self.assertRaises(ValueError) as context:
            Partition(segments, full_set)
        self.assertIn('Each segment must be of type "set".', str(context.exception))

    def test_invalid_segment_item_type(self):
        """Test for a segment item that is not of type 'int'."""
        segments = [{1, '2'}, {3, 4, 5}]
        full_set = {1, 2, 3, 4, 5}
        with self.assertRaises(ValueError) as context:
            Partition(segments, full_set)
        self.assertIn(
            'Each item in the segment must be an integer.', str(context.exception)
        )

    def test_intersection_between_segments(self):
        """Test for segments that have an intersection."""
        segments = [{1, 2, 3}, {3, 4, 5}]
        full_set = {1, 2, 3, 4, 5}
        with self.assertRaises(ValueError) as context:
            Partition(segments, full_set)
        self.assertIn('Elements in two different segments:', str(context.exception))

    def test_union_does_not_match_full_set(self):
        """Test for segments whose union does not match the full set."""
        segments = [{1, 2}, {4, 5}]
        full_set = {1, 2, 3, 4, 5}
        with self.assertRaises(ValueError) as context:
            Partition(segments, full_set)
        self.assertIn(
            'The union of segments does not match the given set.',
            str(context.exception),
        )
        self.assertIn('Missing elements:', str(context.exception))

    def test_iterator(self):
        """Test the iterator of the Partition class."""
        segments = [{1, 2}, {3, 4, 5}]
        full_set = {1, 2, 3, 4, 5}
        partition = Partition(segments, full_set)

        # Verify the order and content of the yielded segments
        for expected_segment, yielded_segment in zip(segments, partition):
            self.assertEqual(expected_segment, yielded_segment)


if __name__ == '__main__':
    unittest.main()



import pandas as pd

from biogeme.audit_tuple import AuditTuple
from biogeme.database import Database
from biogeme.default_parameters import MISSING_VALUE
from biogeme.expressions import (
    Expression,
    ExpressionOrNumeric,
    PanelLikelihoodTrajectory,
    Variable,
    list_of_variables_in_expression,
)

CHOICE_LABEL = 'Choice'
AVAILABILITY_LABEL = 'Avail. '


def audit_variables(expression: Expression, database: Database) -> AuditTuple:

    all_variables: list[Variable] = list_of_variables_in_expression(expression)
    list_of_errors = []
    list_of_warnings = []
    for variable in all_variables:
        if variable.name not in database.dataframe.columns:
            error_msg = f'Variable "{variable.name}" not found in the database. Available variables: {database.dataframe.columns}'
            list_of_errors.append(error_msg)

    return AuditTuple(errors=list_of_errors, warnings=list_of_warnings)


def audit_panel(expression: Expression, database: Database) -> AuditTuple:
    list_of_errors = []
    list_of_warnings = []
    if not database.is_panel():
        return AuditTuple(errors=list_of_errors, warnings=list_of_warnings)

    all_variables: list[Variable] = list_of_variables_in_expression(expression)
    if all_variables and not expression.embed_expression(PanelLikelihoodTrajectory):
        error_msg = (
            f'Expression {expression} does not contain  "PanelLikelihoodTrajectory" although the data has been '
            f'declared to have a panel structure.'
        )
        list_of_errors.append(error_msg)
    return AuditTuple(errors=list_of_errors, warnings=list_of_warnings)


def audit_chosen_alternative(
    choice: ExpressionOrNumeric,
    availability: dict[int, ExpressionOrNumeric],
    database: Database,
) -> AuditTuple:
    from .model_elements import ModelElements
    from biogeme.calculator import MultiRowEvaluator

    """Checks all the rows in the database such that the chosen alternative is not available"""

    list_of_errors = []

    dict_of_expressions = {CHOICE_LABEL: choice} | {
        f'{AVAILABILITY_LABEL}{alt_id:.1f}': the_expression
        for alt_id, the_expression in availability.items()
    }
    model_elements = ModelElements(expressions=dict_of_expressions, database=database)

    the_evaluator: MultiRowEvaluator = MultiRowEvaluator(
        model_elements=model_elements, numerically_safe=True
    )
    results: pd.DataFrame = the_evaluator.evaluate({})

    def chosen_unavailable(row):
        choice_id = row[CHOICE_LABEL]
        if choice_id == MISSING_VALUE:
            return False
        return row[f'{AVAILABILITY_LABEL}{choice_id:.1f}'] == 0

    invalid_indices = results.apply(chosen_unavailable, axis=1)
    problematic_rows = results[invalid_indices]

    list_of_warnings = [
        f'Row index {idx}: chosen alternative {row[CHOICE_LABEL]} is not available'
        for idx, row in problematic_rows.iterrows()
    ]
    return AuditTuple(errors=list_of_errors, warnings=list_of_warnings)



"""

biogeme.tools
=============

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Sat Dec  2 13:09:42 2023
"""

import numpy as np
import pandas as pd

from biogeme.function_output import FunctionOutput
from biogeme.version import getText
from biogeme import tools
import biogeme.biogeme_logging as blog
import biogeme.exceptions as excep

# %%
# Version of Biogeme.
print(getText())

# %%
logger = blog.get_screen_logger(level=blog.INFO)


# %%
# Define a function and its derivatives:
#
# .. math:: f = \log(x_0) + \exp(x_1),
#
# .. math:: g = \left( \begin{array}{c}\frac{1}{x_0} \\
#     \exp(x_1)\end{array}\right),
#
# .. math:: h=\left(\begin{array}{cc} -\frac{1}{x_0^2} & 0 \\ 0 &
#     \exp(x_1)\end{array}\right).
def my_function(x: np.ndarray) -> FunctionOutput:
    """Implementation of the test function.

    :param x: point at which the function and its derivatives must be evaluated.
    """
    f = np.log(x[0]) + np.exp(x[1])
    g = np.empty(2)
    g[0] = 1.0 / x[0]
    g[1] = np.exp(x[1])
    H = np.empty((2, 2))
    H[0, 0] = -1.0 / x[0] ** 2
    H[0, 1] = 0.0
    H[1, 0] = 0.0
    H[1, 1] = np.exp(x[1])
    return FunctionOutput(function=f, gradient=g, hessian=H)


# %%
# Evaluate the function at the point
#
# .. math:: x = \left( \begin{array}{c}1.1 \\ 1.1 \end{array}\right).
#
x = np.array([1.1, 1.1])
the_output = my_function(x)


# %%
the_output.function

# %%
# We use the `DataFrame` for a nicer display.
pd.DataFrame(the_output.gradient)

# %%
pd.DataFrame(the_output.hessian)

# %%
# Calculates an approximation of the gradient by finite differences.
g_fd = tools.findiff_g(my_function, x)

# %%
pd.DataFrame(g_fd)

# %%
# Check the precision of the approximation
pd.DataFrame(the_output.gradient - g_fd)

# %%
# Calculates an approximation of the Hessian by finite differences.
H_fd = tools.findiff_H(my_function, x)

# %%
pd.DataFrame(H_fd)

# %%
# Check the precision of the approximation
pd.DataFrame(the_output.hessian - H_fd)

# %%
# There is a function that checks the analytical derivatives by
# comparing them to their finite difference approximation.
f, g, h, gdiff, hdiff = tools.checkDerivatives(my_function, x, names=None, logg=True)

# %%
# To help reading the reporting, it is possible to give names to variables.

# %%
f, g, h, gdiff, hdiff = tools.checkDerivatives(
    my_function, x, names=['First', 'Second'], logg=True
)

# %%
pd.DataFrame(gdiff)

# %%
hdiff

# %%
# Prime numbers: calculate prime numbers lesser or equal to an upper bound.
myprimes = tools.calculate_prime_numbers(10)
myprimes

# %%
myprimes = tools.calculate_prime_numbers(100)
myprimes

# %%
# Calculate a given number of prime numbers.
myprimes = tools.get_prime_numbers(7)
myprimes

# %%
# Counting groups of data.
alist = [1, 2, 2, 3, 3, 3, 4, 1, 1]

# %%
df = pd.DataFrame(
    {
        'ID': [1, 1, 2, 3, 3, 1, 2, 3],
        'value': [1000, 2000, 3000, 4000, 5000, 5000, 10000, 20000],
    }
)

# %%
tools.countNumberOfGroups(df, 'ID')

# %%
tools.countNumberOfGroups(df, 'value')

# %%
# Likelihood ratio test.
model1 = (-1340.8, 5)
model2 = (-1338.49, 7)

# %%
# A likelihood ratio test is performed. The function returns the
# outcome of the test, the statistic, and the threshold.
tools.likelihood_ratio_test(model1, model2)

# %%
# The default level of significance is 0.95. It can be changed.
tools.likelihood_ratio_test(model1, model2, significance_level=0.9)

# %%
# The order in which the models are presented is irrelevant.
tools.likelihood_ratio_test(model2, model1)

# %%
# But the unrestricted model must have a higher loglikelihood than the
# restricted one.
model1 = (-1340.8, 7)
model2 = (-1338.49, 5)

# %%
try:
    tools.likelihood_ratio_test(model1, model2)
except excep.BiogemeError as e:
    print(e)



import logging
from typing import NamedTuple
from itertools import chain

from biogeme.exceptions import BiogemeError

logger = logging.getLogger(__name__)


class AuditTuple(NamedTuple):
    errors: list[str]
    warnings: list[str]


def merge_audit_tuples(audit_tuples: list[AuditTuple]) -> AuditTuple:
    merged_errors = list(chain.from_iterable(a.errors for a in audit_tuples))
    merged_warnings = list(chain.from_iterable(a.warnings for a in audit_tuples))
    return AuditTuple(errors=merged_errors, warnings=merged_warnings)


def display_messages(audit_tuple: AuditTuple) -> None:
    for warning in audit_tuple.warnings:
        if warning:
            logger.warning(warning)
    if audit_tuple.errors:
        raise BiogemeError(audit_tuple.errors)



"""
Possible approximations of the variance-covariance matrix
Michel Bierlaire
Mon Jun 23 2025, 17:21:45
"""

from enum import Enum


class EstimateVarianceCovariance(str, Enum):
    """Identifies the estimate of the variance-covariance matrix to be used."""

    RAO_CRAMER = "Rao-Cramer"
    ROBUST = "Robust"
    BOOTSTRAP = "Bootstrap"
    BHHH = "BHHH"

    def __str__(self) -> str:
        return self.value



import unittest
import pandas as pd
from datetime import timedelta


from biogeme.results_processing import (
    EstimationResults,
    RawEstimationResults,
    get_pandas_one_parameter,
    get_pandas_estimated_parameters,
    get_pandas_one_pair_of_parameters,
    get_pandas_correlation_results,
    EstimateVarianceCovariance,
)


class TestPandasGeneration(unittest.TestCase):

    def setUp(self):
        # Setting up real estimation results using RawEstimationResults
        raw_estimation = RawEstimationResults(
            model_name='test_model',
            user_notes='user notes',
            beta_names=['beta1', 'beta2'],
            beta_values=[1.2, 1.3],
            lower_bounds=[-10, -11],
            upper_bounds=[None, 23],
            gradient=[0.001, -0.002],
            hessian=[[1, 2], [3, 4]],
            bhhh=[[10, 20], [30, 40]],
            null_log_likelihood=-1000,
            initial_log_likelihood=-2000,
            final_log_likelihood=-500,
            data_name='test_data',
            sample_size=1234,
            number_of_observations=987,
            monte_carlo=True,
            number_of_draws=10_000,
            types_of_draws={'var1': 'UNIFORM'},
            number_of_excluded_data=10,
            draws_processing_time=timedelta(seconds=10),
            optimization_messages={'Diagnostic': 'test'},
            convergence=True,
            bootstrap=[[1.0, 1.0], [2.0, 2.0]],
            bootstrap_time=timedelta(seconds=20),
        )

        self.estimation_results = EstimationResults(
            raw_estimation_results=raw_estimation
        )

    def test_get_pandas_one_parameter(self):
        # Test for getting one parameter in pandas format
        parameter_index = 0
        result = get_pandas_one_parameter(
            estimation_results=self.estimation_results,
            parameter_index=parameter_index,
            variance_covariance_type=EstimateVarianceCovariance.ROBUST,
        )

        # Check that the result is a dictionary and contains expected keys
        self.assertIsInstance(result, dict)
        self.assertIn('Name', result)
        self.assertIn('Value', result)
        self.assertIn('Robust std err.', result)
        self.assertIn('Robust t-stat.', result)
        self.assertIn('Robust p-value', result)

        # Validate parameter values
        self.assertEqual(result['Name'], 'beta1')
        self.assertEqual(result['Value'], 1.2)

    def test_get_pandas_estimated_parameters(self):
        # Test for getting all estimated parameters in pandas DataFrame format
        df = get_pandas_estimated_parameters(
            estimation_results=self.estimation_results,
            variance_covariance_type=EstimateVarianceCovariance.ROBUST,
        )

        # Check if the result is a DataFrame
        self.assertIsInstance(df, pd.DataFrame)

        # Ensure the DataFrame has expected columns and data
        self.assertIn('Name', df.columns)
        self.assertIn('Value', df.columns)
        self.assertIn('Robust std err.', df.columns)
        self.assertIn('Robust t-stat.', df.columns)
        self.assertIn('Robust p-value', df.columns)

        # Verify the content for the first parameter
        self.assertEqual(df.iloc[0]['Name'], 'beta1')
        self.assertEqual(df.iloc[0]['Value'], 1.2)

    def test_get_pandas_one_pair_of_parameters(self):
        # Test for getting one pair of parameter correlations in pandas format
        first_index = 0
        second_index = 1
        result = get_pandas_one_pair_of_parameters(
            estimation_results=self.estimation_results,
            first_parameter_index=first_index,
            second_parameter_index=second_index,
            variance_covariance_type=EstimateVarianceCovariance.ROBUST,
        )

        # Check that the result is a dictionary and contains expected keys
        self.assertIsInstance(result, dict)
        self.assertIn('First parameter', result)
        self.assertIn('Second parameter', result)
        self.assertIn('Robust covariance', result)
        self.assertIn('Robust correlation', result)
        self.assertIn('Robust t-stat.', result)
        self.assertIn('Robust p-value', result)

        # Validate parameter names and covariance/correlation
        self.assertEqual(result['First parameter'], 'beta1')
        self.assertEqual(result['Second parameter'], 'beta2')

    def test_get_pandas_correlation_results(self):
        # Test for getting the correlation results in pandas DataFrame format
        df = get_pandas_correlation_results(
            estimation_results=self.estimation_results,
            variance_covariance_type=EstimateVarianceCovariance.ROBUST,
        )

        # Check if the result is a DataFrame
        self.assertIsInstance(df, pd.DataFrame)
        # Ensure the DataFrame has expected columns and data
        self.assertIn('First parameter', df.columns)
        self.assertIn('Second parameter', df.columns)
        self.assertIn('Robust covariance', df.columns)
        self.assertIn('Robust correlation', df.columns)
        self.assertIn('Robust t-stat.', df.columns)
        self.assertIn('Robust p-value', df.columns)

        # Validate the content for the first pair of parameters
        self.assertEqual(df.iloc[0]['First parameter'], 'beta2')
        self.assertEqual(df.iloc[0]['Second parameter'], 'beta1')


if __name__ == '__main__':
    unittest.main()



"""
Test the models module

:author: Michel Bierlaire
:date: Fri Jul 21 17:17:37 2023
"""

import unittest

from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta, BinaryMax, BinaryMin, Expression, Variable
from biogeme.models import (
    ordered_logit,
    ordered_probit,
    piecewise_as_variable,
    piecewise_formula,
    piecewise_function,
    piecewise_variables,
)


class TestOrderedLogit(unittest.TestCase):
    def test_two_discrete_values(self):
        result = ordered_logit(
            continuous_value=10,
            list_of_discrete_values=[1, 2],
            tau_parameter=Beta('tau', 0, None, None, 0),
        )
        self.assertIsInstance(result, dict)

    def test_multiple_discrete_values(self):
        result = ordered_logit(
            continuous_value=10,
            list_of_discrete_values=[1, 2, 3, 4],
            tau_parameter=Beta('tau', 0, None, None, 0),
        )
        self.assertIsInstance(result, dict)

    def test_not_a_parameter(self):
        with self.assertRaises(BiogemeError):
            ordered_logit(10, [1, 2, 3], 1)


class TestOrderedProbit(unittest.TestCase):
    def test_two_discrete_values(self):
        result = ordered_probit(
            continuous_value=10,
            list_of_discrete_values=[1, 2],
            tau_parameter=Beta('tau', 0, None, None, 0),
        )
        self.assertIsInstance(result, dict)

    def test_multiple_discrete_values(self):
        result = ordered_probit(
            continuous_value=10,
            list_of_discrete_values=[1, 2, 3, 4],
            tau_parameter=Beta('tau', 0, None, None, 0),
        )
        self.assertIsInstance(result, dict)

    def test_not_a_parameter(self):
        with self.assertRaises(BiogemeError):
            ordered_probit(10, [1, 2, 3], 1)


class TestPiecewiseVariables(unittest.TestCase):
    def test_correct_input(self):
        variable = "x"
        thresholds = [None, 10, 20, None]
        expected_number_of_variables = len(thresholds) - 1
        result = piecewise_variables(variable, thresholds)
        expected_result = [
            BinaryMin(Variable('x'), 10.0),
            BinaryMax(0.0, BinaryMin((Variable('x') - 10.0), 10.0)),
            BinaryMax(0.0, (Variable('x') - 20.0)),
        ]
        self.assertEqual(len(result), expected_number_of_variables)
        for expected, obtained in zip(expected_result, result):
            self.assertEqual(
                repr(expected),
                repr(obtained),
            )

    def test_all_none_thresholds(self):
        variable = "x"
        thresholds = [None, None, None]
        with self.assertRaises(BiogemeError):
            piecewise_variables(variable, thresholds)

    def test_invalid_none_thresholds(self):
        variable = "x"
        thresholds = [None, None, 20, 30]
        with self.assertRaises(BiogemeError):
            piecewise_variables(variable, thresholds)

    def test_wrong_variable_type(self):
        variable = 123  # Not a string or Expression
        thresholds = [10, 20]
        expected_error_msg = "Expression of type Variable expected, not <class 'int'>"
        with self.assertRaisesRegex(BiogemeError, expected_error_msg):
            piecewise_variables(variable, thresholds)

    def test_edge_case_empty_thresholds(self):
        variable = "x"
        thresholds = []
        expected_error_msg = 'No threshold has been provided.'
        with self.assertRaisesRegex(BiogemeError, expected_error_msg):
            piecewise_variables(variable, thresholds)


class TestPiecewiseFormula(unittest.TestCase):

    def test_valid_input_without_betas(self):
        # Test to ensure function returns correct Expression with valid inputs and no betas
        variable = "x"
        thresholds = [None, 10, 20, None]
        result = piecewise_formula(variable, thresholds)
        expected_result = 'MultipleSum'
        self.assertTrue(str(result).startswith(expected_result))

    def test_valid_input_with_betas(self):
        # Test to ensure function returns correct Expression with valid inputs and betas
        variable = "x"
        thresholds = [None, 10, 20, None]
        beta_1 = Beta('beta_1', 0, None, None, 0)
        beta_2 = Beta('beta_2', 0, None, None, 0)
        beta_3 = Beta('beta_3', 0, None, None, 0)
        betas = [beta_1, beta_2, beta_3]
        result = piecewise_formula(
            variable=variable, thresholds=thresholds, betas=betas
        )
        expected_result = 'MultipleSum'
        self.assertTrue(str(result).startswith(expected_result))

    def test_variable_not_variable_or_str(self):
        # Test to check function raises an error if `variable` is neither a Variable instance nor a string
        variable = 123  # Incorrect type
        thresholds = [10, 20]
        with self.assertRaisesRegex(
            BiogemeError,
            'The first argument of piecewiseFormula must be the name of a variable, or the variable itself.',
        ):
            piecewise_formula(variable, thresholds)

    def test_all_none_thresholds(self):
        # Test to check function raises an error if all thresholds are None
        variable = "x"
        thresholds = [None, None, None]
        with self.assertRaisesRegex(
            BiogemeError,
            'All thresholds for the piecewise linear specification are set to None.',
        ):
            piecewise_formula(variable, thresholds)

    def test_invalid_none_in_thresholds(self):
        # Test to check function raises an error if None is used incorrectly in thresholds
        variable = "x"
        thresholds = [None, None, 20, 30]
        with self.assertRaisesRegex(
            BiogemeError, 'only the first and the last thresholds can be None'
        ):
            piecewise_formula(variable, thresholds)

    def test_incorrect_betas_length(self):
        # Test to check function raises an error if the length of betas does not match the expected
        variable = "x"
        thresholds = [None, 10, 20, None]
        beta_1 = Beta('beta_1', 0, None, None, 0)
        beta_2 = Beta('beta_2', 0, None, None, 0)
        betas = [beta_1, beta_2]
        expected_error_msg = 'As there are 4 thresholds, a total of 3 Beta parameters are needed, and not 2.'
        with self.assertRaisesRegex(BiogemeError, expected_error_msg):
            piecewise_formula(variable, thresholds, betas)


class TestPiecewiseAsVariable(unittest.TestCase):

    def test_valid_input_without_betas(self):
        variable = Variable('x')
        thresholds = [None, 10, 20, None]
        # Call the function under test
        result = piecewise_as_variable(variable.name, thresholds)
        # Assert the result is an Expression
        self.assertIsInstance(result, Expression)
        expected_result = '(BinaryMin(x, `10.0`)'
        self.assertTrue(str(result).startswith(expected_result))

    def test_valid_input_with_betas(self):
        variable = 'x'
        thresholds = [None, 10, 20, None]
        beta_1 = Beta('beta_1', 0, None, None, 0)
        beta_2 = Beta('beta_2', 0, None, None, 0)
        betas = [beta_1, beta_2]
        result = piecewise_as_variable(variable, thresholds, betas)
        self.assertIsInstance(result, Expression)
        expected_result = '(BinaryMin(x, `10.0`)'
        self.assertTrue(str(result).startswith(expected_result))

    def test_invalid_variable_type(self):
        variable = 123  # Incorrect type
        thresholds = [10, 20]
        with self.assertRaises(BiogemeError):
            piecewise_as_variable(variable, thresholds)

    def test_all_none_thresholds_error_message(self):
        variable = 'x'
        thresholds = [None, None, None]
        with self.assertRaisesRegex(
            BiogemeError,
            'All thresholds for the piecewise linear specification are set to None.',
        ):
            piecewise_as_variable(variable, thresholds)

    def test_incorrect_betas_length_error_message(self):
        variable = 'x'
        thresholds = [None, 10, 20, None]
        betas = [
            Beta(f'beta_{variable}_minus_inf_10', 0, None, None, 0)
        ]  # Incorrect number
        with self.assertRaisesRegex(
            BiogemeError, 'a total of 2 Beta parameters are needed, and not 1.'
        ):
            piecewise_as_variable(variable, thresholds, betas)


class TestPiecewiseFunction(unittest.TestCase):

    def test_single_interval(self):
        x = 5
        thresholds = [None, 10]  # -inf to 10
        betas = [1]  # Only one beta, as there's only one interval
        result = piecewise_function(x, thresholds, betas)
        self.assertEqual(
            result,
            5,
            "Should match the input x because it's within the interval and beta=1.",
        )

    def test_multiple_intervals(self):
        x = 15
        thresholds = [None, 10, 20, None]  # -inf to 10, 10 to 20, 20 to inf
        betas = [1, 2, 0.5]  # Different betas for each interval
        result = piecewise_function(x, thresholds, betas)
        self.assertEqual(
            10 + 5 * 2, result, "Should accumulate values across intervals correctly."
        )

    def test_below_first_threshold(self):
        x = -5
        thresholds = [0, 10]  # 0 to 10
        betas = [1]  # Only one beta, as there's only one interval
        result = piecewise_function(x, thresholds, betas)
        self.assertEqual(result, 0, "Should be 0 as x is below the first threshold.")

    def test_above_last_threshold(self):
        x = 25
        thresholds = [0, 10, 20, None]  # 0 to 10, 10 to 20, 20 to inf
        betas = [1, 2, 3]  # Different betas for each interval
        result = piecewise_function(x, thresholds, betas)
        self.assertEqual(
            result,
            10 + 20 + 15,
            "Should handle values above the last threshold correctly.",
        )

    def test_all_none_thresholds(self):
        x = 5
        thresholds = [None, None]  # Invalid configuration
        betas = [1]
        with self.assertRaisesRegex(
            BiogemeError,
            'All thresholds for the piecewise linear specification are set to None.',
        ):
            piecewise_function(x, thresholds, betas)

    def test_incorrect_betas_length(self):
        x = 5
        thresholds = [None, 10, 20, None]
        betas = [1, 2]  # Incorrect number of betas
        with self.assertRaisesRegex(
            BiogemeError,
            'a total of 3 values are needed to initialize the parameters. But 2 are provided',
        ):
            piecewise_function(x, thresholds, betas)

    def test_invalid_none_in_thresholds(self):
        x = 5
        thresholds = [0, None, 20]  # Invalid configuration
        betas = [1, 2]
        with self.assertRaisesRegex(
            BiogemeError,
            'For piecewise linear specification, only the first and the last thresholds can be None',
        ):
            piecewise_function(x, thresholds, betas)


if __name__ == '__main__':
    unittest.main()



"""Arithmetic expressions accepted by Biogeme: belongs to

Michel Bierlaire
Sat May 03 2025, 11:56:33
"""

from __future__ import annotations

import logging

import jax.numpy as jnp

from .base_expressions import ExpressionOrNumeric
from .jax_utils import JaxFunctionType
from .unary_expressions import UnaryOperator

logger = logging.getLogger(__name__)


class BelongsTo(UnaryOperator):
    """
    Check if a value belongs to a set
    """

    def __init__(self, child: ExpressionOrNumeric, the_set: set[float]):
        """Constructor

        :param child: arithmetic expression
        :type child: biogeme.expressions.Expression
        :param the_set: set of values
        :type the_set: set(float)
        """
        super().__init__(child)
        self.the_set: set[float] = the_set

    def deep_flat_copy(self) -> BelongsTo:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        child_copy = self.child.deep_flat_copy()
        return type(self)(child=child_copy, the_set=self.the_set)

    def __str__(self) -> str:
        return f'BelongsTo({self.child}, "{self.the_set}")'

    def __repr__(self) -> str:
        return f'BelongsTo({self.child}, "{self.the_set}")'

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        child_jax = self.child.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            child_value = child_jax(
                parameters, one_row, the_draws, the_random_variables
            )

            return jnp.where(
                jnp.isin(child_value, jnp.array(list(self.the_set))), 1.0, 0.0
            )

        return the_jax_function



from typing import NamedTuple

from scipy.stats import chi2

from biogeme.exceptions import BiogemeError


class LRTuple(NamedTuple):
    """Tuple for the likelihood ratio test"""

    message: str
    statistic: float
    threshold: float

    def __str__(self):
        return f'{self.message} (statistic: {self.statistic:.2f}, threshold: {self.threshold:.2f})'


def likelihood_ratio_test(
    model1: tuple[float, int],
    model2: tuple[float, int],
    significance_level: float = 0.05,
) -> LRTuple:
    """This function performs a likelihood ratio test between a
    restricted and an unrestricted model.

    :param model1: the final loglikelihood of one model, and the number of
                   estimated parameters.

    :param model2: the final loglikelihood of the other model, and
                   the number of estimated parameters.

    :param significance_level: level of significance of the test. Default: 0.05

    :return: a tuple containing:

                  - a message with the outcome of the test
                  - the statistic, that is minus two times the difference
                    between the loglikelihood  of the two models
                  - the threshold of the chi square distribution.

    :raise BiogemeError: if the unrestricted model has a lower log
        likelihood than the restricted model.

    """

    log_like_m1, df_m1 = model1
    log_like_m2, df_m2 = model2
    if log_like_m1 > log_like_m2:
        if df_m1 < df_m2:
            raise BiogemeError(
                f'The unrestricted model {model2} has a '
                f'lower log likelihood than the restricted one {model1}'
            )
        log_like_ur = log_like_m1
        log_like_r = log_like_m2
        df_ur = df_m1
        df_r = df_m2
    else:
        if df_m1 >= df_m2:
            raise BiogemeError(
                f'The unrestricted model {model1} has a '
                f'lower log likelihood than the restricted one {model2}'
            )
        log_like_ur = log_like_m2
        log_like_r = log_like_m1
        df_ur = df_m2
        df_r = df_m1

    stat = -2 * (log_like_r - log_like_ur)
    chi_df = df_ur - df_r
    threshold = chi2.ppf(1 - significance_level, chi_df)
    if stat <= threshold:
        final_msg = f'H0 cannot be rejected at level {100*significance_level:.1f}%'
    else:
        final_msg = f'H0 can be rejected at level {100*significance_level:.1f}%'
    return LRTuple(message=final_msg, statistic=stat, threshold=threshold)



"""

Nested logit model
==================

Example of a nested logit model, using the original syntax for nests.
Since biogeme 3.13, a new syntax, more explicit, has been adopted.

:author: Michel Bierlaire, EPFL
:date: Tue Oct 24 13:37:27 2023

"""

import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b09nested')

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
MU = Beta('MU', 1, 1, 10, 0)

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of nests. In this example, we create a nest for the
# existing modes, that is train (1) and car (3).  Each nest is
# associated with a tuple containing (i) the nest parameter and (ii)
# the list of alternatives.
existing = MU, [1, 3]
future = 1.0, [2]
nests = existing, future

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
# The choice model is a nested logit, with availability conditions.
logprob = models.lognested(V, av, nests, CHOICE)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = "b09nested"

# %%
# Calculate the null log likelihood for reporting.
the_biogeme.calculateNullLoglikelihood(av)

# %%
# Estimate the parameters.
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
pandas_results = results.getEstimatedParameters()
pandas_results



"""Representation of unknown parameters

:author: Michel Bierlaire
:date: Sat Apr 20 14:54:16 2024

"""

from __future__ import annotations

import logging

import jax.numpy as jnp
from biogeme.exceptions import BiogemeError

from .elementary_expressions import Elementary
from .elementary_types import TypeOfElementaryExpression
from .jax_utils import JaxFunctionType

logger = logging.getLogger(__name__)


class Beta(Elementary):
    """
    Unknown parameters to be estimated from data.
    """

    def __init__(
        self,
        name: str,
        value: float,
        lowerbound: float | None,
        upperbound: float | None,
        status: int,
    ):
        """Constructor

        :param name: name of the parameter.
        :param value: default value.
        :param lowerbound: if different from None, imposes a lower
          bound on the value of the parameter during the optimization.
        :param upperbound: if different from None, imposes an upper
          bound on the value of the parameter during the optimization.
        :param status: if different from 0, the parameter is fixed to
          its default value, and not modified by the optimization algorithm.

        :raise BiogemeError: if the first parameter is not a str.

        :raise BiogemeError: if the second parameter is not an int or a float.
        """

        if not isinstance(value, (int, float)):
            error_msg = (
                f"The second parameter for {name} must be "
                f"a float and not a {type(value)}: {value}"
            )
            raise BiogemeError(error_msg)
        if not isinstance(name, str):
            error_msg = (
                f"The first parameter must be a string and "
                f"not a {type(name)}: {name}"
            )
            raise BiogemeError(error_msg)
        super().__init__(name)
        self.init_value = value
        self.lower_bound = lowerbound
        self.upper_bound = upperbound
        self.status = status

    def deep_flat_copy(self) -> Beta:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        return type(self)(
            name=self.name,
            value=self.init_value,
            lowerbound=self.lower_bound,
            upperbound=self.upper_bound,
            status=self.status,
        )

    @property
    def is_free(self):
        return self.status == 0

    @property
    def is_fixed(self):
        return not self.is_free

    @property
    def expression_type(self) -> TypeOfElementaryExpression:
        """Type of elementary expression"""
        if self.is_free:
            return TypeOfElementaryExpression.FREE_BETA
        return TypeOfElementaryExpression.FIXED_BETA

    def __str__(self) -> str:
        return (
            f"Beta('{self.name}', {self.init_value}, {self.lower_bound}, "
            f"{self.upper_bound}, {self.status})"
        )

    def __repr__(self) -> str:
        return f'<Beta name={self.name} value={self.init_value} status={self.status}>'

    def get_value(self) -> float:
        """Calculates the value of the expression if it is simple"""
        return self.init_value

    @property
    def safe_beta_id(self) -> int:
        """Check the presence of the ID before using it"""
        if self.specific_id is None:
            raise BiogemeError(f"No id defined for parameter {self.name}")
        return self.specific_id

    def fix_betas(
        self,
        beta_values: dict[str, float],
        prefix: str | None = None,
        suffix: str | None = None,
    ):
        """Fix all the values of the Beta parameters appearing in the
        dictionary

        :param beta_values: dictionary containing the betas to be
            fixed (as key) and their value.
        :type beta_values: dict(str: float)

        :param prefix: if not None, the parameter is renamed, with a
            prefix defined by this argument.
        :type prefix: str

        :param suffix: if not None, the parameter is renamed, with a
            suffix defined by this argument.
        :type suffix: str

        """
        if self.name in beta_values:
            self.init_value = beta_values[self.name]
            self.status = 1
            if prefix is not None:
                self.name = f"{prefix}{self.name}"
            if suffix is not None:
                self.name = f"{self.name}{suffix}"

    def change_init_values(self, betas: dict[str, float]):
        """Modifies the initial values of the Beta parameters.

        The fact that the parameters are fixed or free is irrelevant here.

        :param betas: dictionary where the keys are the names of the
                      parameters, and the values are the new value for
                      the parameters.
        :type betas: dict(string:float)
        """

        value = betas.get(self.name)
        if value is not None and value != self.init_value:
            if self.is_fixed:
                warning_msg = (
                    f'Parameter {self.name} is fixed, but its value '
                    f'is changed from {self.init_value} to {value}.'
                )
                logger.warning(warning_msg)
            self.init_value = value

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Returns a compiled JAX-compatible function that extracts the beta value from the parameter vector using its
        unique index.
        """
        if self.is_free:

            def the_jax_function(
                parameters: jnp.ndarray,
                one_row: jnp.ndarray,
                the_draws: jnp.ndarray,
                the_random_variables: jnp.ndarray,
            ) -> jnp.array:
                return jnp.asarray(parameters[self.safe_beta_id])

            return the_jax_function

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.array:
            return jnp.asarray(self.init_value)

        return the_jax_function



import time
from datetime import timedelta
from typing import Callable, Any


def format_timedelta(td: timedelta) -> str:
    """Format a timedelta in a "human-readable" way"""

    # Determine the total amount of seconds
    total_seconds = int(td.total_seconds())
    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    # Get the total microseconds remaining
    microseconds = td.microseconds

    # Format based on the most significant unit
    if hours > 0:
        return f'{hours}h {minutes}m {seconds}s'
    if minutes > 0:
        return f'{minutes}m {seconds}s'
    if seconds > 0:
        return f'{seconds}.{microseconds // 100000:01}s'
    if microseconds >= 1000:
        return f'{microseconds // 1000}ms'  # Convert to milliseconds

    return f'{microseconds}μs'  # Microseconds as is


class Timing:
    """Call for timing the execution of callable functions"""

    def __init__(self, warm_up_runs: int = 10, num_runs: int = 100):
        self._warm_up_runs = warm_up_runs
        self._num_runs = num_runs

    @property
    def warm_up_runs(self) -> int:
        return self._warm_up_runs

    @warm_up_runs.setter
    def warm_up_runs(self, value: int):
        if value < 0:
            raise ValueError("warm_up_runs must be non-negative")
        self._warm_up_runs = value

    @property
    def num_runs(self) -> int:
        return self._num_runs

    @num_runs.setter
    def num_runs(self, value: int):
        if value < 1:
            raise ValueError("num_runs must be at least 1")
        self._num_runs = value

    def time_function(
        self,
        callable_func: Callable,
        kwargs: dict[str, Any],
    ) -> float:
        """
        Times the execution of a callable function, excluding the warm-up period.

        :param callable_func: The function to be timed.
        :param kwargs: Dictionary of keyword arguments to pass to the function. Note that all arguments must be with
            keyword
        :return: Average time per run.
        """

        # Warm-up period
        for i in range(self.warm_up_runs):
            callable_func(**kwargs)

        # Timing the function
        start_time = time.time()
        for _ in range(self.num_runs):
            callable_func(**kwargs)
        end_time = time.time()

        # Calculate the average time per run
        total_time = end_time - start_time
        average_time_per_run = total_time / self.num_runs

        return average_time_per_run



"""

Simulation of a choice model
============================

We use an estimated model to perform various simulations.

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:04:33 2023

"""
import sys
import time
import pandas as pd
from biogeme import models
import biogeme.biogeme as bio
import biogeme.exceptions as excep
import biogeme.results as res
from optima_data import database, normalized_weight
from scenarios import scenario

# %%
# Obtain the specification for the default scenario.
# The definition of the scenarios is available in :ref:`scenarios`.
V, nests, _, _ = scenario()

V_PT = V[0]
V_CAR = V[1]
V_SM = V[2]

# %%
# Obtain the expression for the choice probability of each alternative.
prob_PT = models.nested(V, None, nests, 0)
prob_CAR = models.nested(V, None, nests, 1)
prob_SM = models.nested(V, None, nests, 2)

# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='saved_results/b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02simulation.py '
        'in order to generate the '
        'file b02estimation.pickle.'
    )

# %%
# We now simulate various expressions on the database, and store the
# results in a Pandas dataframe.

# %%
start_time = time.time()
simulate_formulas = {
    'weight': normalized_weight.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Utility PT': V_PT.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Utility car': V_CAR.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Utility SM': V_SM.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Prob. PT': prob_PT.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Prob. car': prob_CAR.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
    'Prob. SM': prob_SM.getValue_c(
        betas=results.getBetaValues(), database=database, prepareIds=True
    ),
}

# %%
simulated_values = pd.DataFrame.from_dict(
    simulate_formulas,
)

# %%
print(
    f'--- Execution time with getValue_c: '
    f'{time.time() - start_time:.2f} seconds ---'
)

# %%
# We now perform the same simulation using Biogeme. The results are
# identical, but the execution time is faster. Indeed, Biogeme
# recycles calculations performed for one expression for the other
# expressions.

# %%
# A dictionary with the requested expression must be provided to Biogeme
simulate = {
    'weight': normalized_weight,
    'Utility PT': V_PT,
    'Utility car': V_CAR,
    'Utility SM': V_SM,
    'Prob. PT': prob_PT,
    'Prob. car': prob_CAR,
    'Prob. SM': prob_SM,
}

# %%
start_time = time.time()
the_biogeme = bio.BIOGEME(database, simulate)
biogeme_simulation = the_biogeme.simulate(results.getBetaValues())

# %%
print(
    f'--- Execution time with Biogeme:    '
    f'{time.time() - start_time:.2f} seconds ---'
)

# %%
# Let's print the two results, to show that they are identical

# %%
# Without Biogeme
print(simulated_values)

# %%
# With Biogeme
print(biogeme_simulation)
